import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import csv
import sys
from scipy.optimize import curve_fit

# =============================================================================
# This code contains the functions to take the data base file of chemical
# reactions and put them in SUIB input file format. This functions here are
# either used in db2fort_input.py or in the jupyter notebook.
# =============================================================================
#

data = []


with open('ABINCODES','r') as fin:
    data=fin.readlines()

data = [d.split() for d in data]
abin = [int(d[0]) for d in data]
names = [d[1] + d[2] for d in data]

Nabin = len(data)

# open the mbin cone file
with open('MBINCODES','r') as fin:
    data=fin.readlines()

data = [d.split() for d in data]
mbin = [int(d[0]) for d in data]
mbin_names = [d[1] + d[2] for d in data]
mbin_mass = [d[3] for d in data]


def return_abin(name):
    '''
    This function takes the name of an element or a molecule and returns the 
    numerical code (ABIN), which is used internally as identifiers in the SUMO
    code. 
    
    Parameters
    ----------
    name: string
              The name of the element or molecule.
    Returns
    -------
    abin: int
              ABIN code.
    '''
    indi = names.index(name)
    return abin[indi]

def return_mbin(name):
    '''
    This function takes the name of a molecule and returns the 
    numerical code (MBIN), which is used internally as identifiers in the SUMO
    code. 
    
    Parameters
    ----------
    name: string
              The name of the element or molecule.
    Returns
    -------
    mbin: int
              MBIN code.
    '''
    indi = mbin_names.index(name)
    return Nabin + mbin[indi]

def return_bincode(name):
    try:
        temp = int(name)
        return temp
    except:
        try:
            if name == 'xI':
                return 0
            elif len(name.split('I')[0]) < 3:
                temp = return_abin(name)
            else:
                temp = int(name)
        except:
            temp = return_mbin(name)
        return temp


def get_digits(str1):
    '''
    This function returns the digits of a string.
    
    Parameters
    ----------
    str1: string
              The input string
    Returns
    -------
    : str
            The numbers in the input string
          '''
    return ''.join(c for c in str1 if c.isdigit())


def _sop(x, a, b, c, d):
    '''
    This a function in the form of 
    -
    a * t^b (1 + c * 10^(d * t))
    -
    used to fit time-dependent rates. This is of the same form as used in the
    paper Kingdon and Ferland 1996. 
    
    Parameters
    ----------
    x: float or array
                The independent variable(s).
    a,b,c,d: float
                The equation constants
    Returns
    -------
    temp: float or array
                The output of the defined function.
    '''
    
    temp = a * (x/1e4)**b * (1 + c*np.exp(d * (x/1e4)))
    return temp


    

def suib_input_writer(db_name, sav_name, plot=False):
    '''
    This function imports the human readable data base, and converts
    it into a format that can be read by the SUIB subroutine. For time-dependent 
    reactions a equation is fitted (defined in _sop()). 
        
    Parameters
    ----------
    db_name: string
                The file name of the human readable data base.
    f_name: string
                The name of the SUIB input file.
    plot: bool
                If the fitting should be saved in plots (in /res/)
    Returns
    -------
    None    
    '''
    # imports the data base
    df = pd.read_csv(db_name)
    
    # define the output array
    outp = []

    # iterates over all the reactions in the database
    for index, row in df.iterrows():

        tmax = row['max temp']
        tmin = row['min temp']
            
        tmin = 10. if tmin < 101. else tmin
        
        r1 = return_bincode(row['Reactant 1'])
        r2 = return_bincode(row['Reactant 2'])
        p1 = return_bincode(row['Product 1'])
        p2 = return_bincode(row['Product 2'])
        p3 = return_bincode(row['Product 3'])
            
        temp = [r1, r2, p1, p2, p3, row['alpha'],row['beta'],row['gamma'], tmin, tmax]
        
        # The output is save in the array outp
        outp.append(temp)
    
    # the outp array is saved in an csv file
    with open(sav_name, mode='w') as out_file:
        out_write = csv.writer(out_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        out_write.writerow([len(outp), len(outp[0])])
    
        [out_write.writerow(o) for o in outp]
    
    plt.close('all')
    
    print('The database is now saved in %s, for SUIB to read from. '%sav_name)
    
#    print(outp)
    
    all_rp = return_bincode(row['Reactant 1']) + return_bincode(row['Reactant 2']) + return_bincode(row['Product 1']) + return_bincode(row['Product 2'])
#    print(all_rp)
    

#creates the mr network files
#suib_input_writer('hr_network/hr_umistnetwork_co_newra.csv', 'mr_network/mr_umistnetwork_co_109_newra.csv')

#suib_input_writer('hr_network/hr_umistnetwork_allmol.csv', 'mr_network/mr_umistnetwork_allmol_109.csv')

#suib_input_writer('hr_network/hr_umistnetwork_co.csv', 'mr_network/mr_umistnetwork_co_111.csv')

suib_input_writer('hr_network/hr_umistnetwork_allmol3.csv', 'mr_network/mr_umistnetwork_allmol3_126.csv')



fname  = 'MBINCODES'
data = []
with open(fname, 'r') as fin:
    data=fin.readlines()

outp = []
for d in data:
    d = d.strip()
    d = d.split()
    mb = int(d[0])
    name = d[1]
    
    if d[2] == 'I':
        charge = 0.
    else:
        charge = 1.
    
    mass = d[3]
    
    if any(n.isdigit() for n in name):
        indi = [x.isdigit() for x in name].index(True)
        if indi == len(name)-1:
            n1 = 1.; n2 = 2.
            tname = name[:-1]
            if len(tname) == 1:
                n2 = 0.
                n1 = float(name[indi])
                p1 = tname
                p2 = 'x'
            elif tname[1].islower():
                p1 = tname[:2]
                p2 = tname[2:]
            else:
                p1 = tname[0]
                p2 = tname[1:]
        else:
            n1 = 2.; n2 = 1.
            p1 = name[:indi]
            p2 = name[indi+1:]
    else:
        n1 = 1.; n2 = 1.
        if name[1].islower():
            p1 = name[:2]
            p2 = name[2:]
        else:
            p1 = name[0]
            p2 = name[1:]
    
    
    outp.append([(name+d[2]).ljust(7), mb, return_bincode(p1+'I'), return_bincode(p2+'I'), n1, n2, mass,charge])


with open('moldata_input.csv', 'w', newline='') as file:
    header = ['name', 'mbin', 'parent1', 'parent2', 'n_p1', 'n_p2', 'mass', 'charge']
    mywriter = csv.writer(file, delimiter=',')
    mywriter.writerow(header)
    mywriter.writerows(outp)































