import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import csv
import sys
from scipy.optimize import curve_fit

# =============================================================================
# This code contains the functions to take the data base file of chemical
# reactions and put them in SUIB input file format. This functions here are
# either used in db2fort_input.py or in the jupyter notebook.
# =============================================================================
#

data = []


with open('ABINCODES','r') as fin:
    data=fin.readlines()

print(data)
data = [d.split() for d in data]
abin = [int(d[0]) for d in data]
names = [d[1] + d[2] for d in data]

Nabin = len(data)

# open the mbin cone file
with open('MBINCODES','r') as fin:
    data=fin.readlines()

data = [d.split() for d in data]
mbin = [int(d[0]) for d in data]
mbin_names = [d[1] + d[2] for d in data]


def return_abin(name):
    '''
    This function takes the name of an element or a molecule and returns the 
    numerical code (ABIN), which is used internally as identifiers in the SUMO
    code. 
    
    Parameters
    ----------
    name: string
              The name of the element or molecule.
    Returns
    -------
    abin: int
              ABIN code.
    '''
    indi = names.index(name)
    return abin[indi]

def return_mbin(name):
    '''
    This function takes the name of a molecule and returns the 
    numerical code (MBIN), which is used internally as identifiers in the SUMO
    code. 
    
    Parameters
    ----------
    name: string
              The name of the element or molecule.
    Returns
    -------
    mbin: int
              MBIN code.
    '''
    indi = mbin_names.index(name)
    return Nabin + mbin[indi]

def return_bincode(name):
    try:
        temp = int(name)
        return temp
    except:
        try:
            temp = return_abin(name)
        except:
            temp = return_mbin(name)
        return temp


def get_digits(str1):
    '''
    This function returns the digits of a string.
    
    Parameters
    ----------
    str1: string
              The input string
    Returns
    -------
    : str
            The numbers in the input string
          '''
    return ''.join(c for c in str1 if c.isdigit())


def _sop(x, a, b, c, d):
    '''
    This a function in the form of 
    -
    a * t^b (1 + c * 10^(d * t))
    -
    used to fit time-dependent rates. This is of the same form as used in the
    paper Kingdon and Ferland 1996. 
    
    Parameters
    ----------
    x: float or array
                The independent variable(s).
    a,b,c,d: float
                The equation constants
    Returns
    -------
    temp: float or array
                The output of the defined function.
    '''
    
    temp = a * (x/1e4)**b * (1 + c*np.exp(d * (x/1e4)))
    return temp


    

def suib_input_writer(db_name, sav_name, plot=False):
    '''
    This function imports the human readable data base, and converts
    it into a format that can be read by the SUIB subroutine. For time-dependent 
    reactions a equation is fitted (defined in _sop()). 
        
    Parameters
    ----------
    db_name: string
                The file name of the human readable data base.
    f_name: string
                The name of the SUIB input file.
    plot: bool
                If the fitting should be saved in plots (in /res/)
    Returns
    -------
    None    
    '''
    # imports the data base
    df = pd.read_csv(db_name)
    
    # define the output array
    outp = []

    # iterates over all the reactions in the database
    for index, row in df.iterrows():
        
        # if the rates are temperature dependent the _sop() is fitted
#         CorF = row['Function or Constant']
#         if 'F' in CorF:
#             corf_out = 'F'
#             tr = row['Rates']
#             tr = tr.replace('exp', 'np.exp')
#
#             tmax = row['max temp']
#             tmin = row['min temp']
#
#             ts_int = tmin
#             te_int = tmax
#             ts_int = 200.
#
#             #temp!
#             #-------------------
#             ts_int = 3000.
#             te_int = 5000.
#
#             #temp! keep this!
# #            if tmax > 20000. and tmin < 18000.:
# #                te_int = 20000.
#
#             # defines the temperature interval used for thte fitting
#             t = np.linspace(ts_int, te_int, 1000)
#             test = tr.split('*')
#             if 'KF96' in CorF:
#
#                 yt = eval(tr)
#
#                 temp1 = tr.split('*')
#                 tempA = eval(temp1[0])
#                 tempB = eval(temp1[3])
#                 tempC = eval(temp1[4].split('(1')[1])
#                 tempD = eval(temp1[5].split('(')[1])
#                 best_vals = [tempA, tempB, tempC, tempD]
#
#             else:
#                 yt = eval(tr)
#
#                 # tries the first input guesses
#                 try:
#                     ig = [1e-10, 2., 5., -5.]
#                     best_vals, covar = curve_fit(_sop, t, yt, p0=ig)
#
#                 # if the first input guest does not converge, a more general guess in used
#                 except:
#                     ig = [1e-10, 0., 0., 0.]
#                     best_vals, covar = curve_fit(_sop, t, yt, p0=ig)
#
#             # if plotting is wanted
#             if plot:
#                 plt.figure()
#                 plt.title(row['Rates'])
#                 plt.plot(t, _sop(t, best_vals[0],best_vals[1],best_vals[2], best_vals[3]) , color = 'black', alpha = 0.7)
#                 plt.plot(t, yt, 'r--')
#
#
#                 t_test = 4000.
#                 esti = _sop(t_test, best_vals[0],best_vals[1],best_vals[2], best_vals[3])
#                 t = 4000.
#                 exa = eval(tr)
#                 print(esti)
#
#                 plt.scatter(t_test, esti, marker = 's')
#                 plt.scatter(t_test, exa, marker = 'o')
#
#                 plt.yscale('log')
# #                plt.xscale('log')
#
#                 plt.savefig('res/%s+%s-%s+%s+%s.png'%(row['Reactant 1'], row['Reactant 2'],row['Product 1'],
#                 row['Product 2'], row['Product 3']))
#
#
#                 # print('%s+%s->%s+%s'%(row['Reactant 1'], row['Reactant 2'],row['Product 1'],row['Product 2']))
#                 # print(exa, esti, tmin, tmax, ts_int, te_int)
#                 # print('----')
#                 #
#             r1 = return_bincode(row['Reactant 1'])
#             r2 = return_bincode(row['Reactant 2'])
#             p1 = return_bincode(row['Product 1'])
#             p2 = return_bincode(row['Product 2'])
#             p3 = return_bincode(row['Product 3'])
#
#             temp = [r1, r2, p1, p2, p3, corf_out, best_vals[0],best_vals[1],best_vals[2],
#              best_vals[3], tmin, tmax, row['securevec']]
#
#
#         # if the rates are constant rates
#         else:
        tmax = row['max temp']
        tmin = row['min temp']
            
        tmin = 10. if tmin < 101. else tmin
            
        r1 = return_bincode(row['Reactant 1'])
        r2 = return_bincode(row['Reactant 2'])
        p1 = return_bincode(row['Product 1'])
        p2 = return_bincode(row['Product 2'])
        p3 = return_bincode(row['Product 3'])
            
        temp = [r1, r2, p1, p2, p3, row['alpha'],row['beta'],row['gamma'], tmin, tmax]
        
        # The output is save in the array outp
        outp.append(temp)
    
    # the outp array is saved in an csv file
    with open(sav_name, mode='w') as out_file:
        out_write = csv.writer(out_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        out_write.writerow([len(outp), len(outp[0])])
    
        [out_write.writerow(o) for o in outp]
    
    plt.close('all')
    
    print('The database is now saved in %s, for SUIB to read from. '%sav_name)
    
    print(outp)
    
    all_rp = return_bincode(row['Reactant 1']) + return_bincode(row['Reactant 2']) + return_bincode(row['Product 1']) + return_bincode(row['Product 2'])
    print(all_rp)
    
    
def _fcons():
    '''
    This function defines the atom/molecules conservation
    used in the getf.f90 function. By default the second
    ionised nucleus is the conserved quantity, and XI, XII and
    XII from ABINCODES, as well as all molecules defined in the 
    molecule network is used. The result is saved in nuccons_input.csv. 

    Parameters
    ----------
    mol_list: array of strings
                List of all the molecules included in the mol database
    Returns
    -------
    --




    '''


    # All molecules in MBINCODES
    all_mol = mbin_names

    all_nuc = []

    # names of the molecules
    for a in names:
        nuc = ''
        # Finds the individual nucleus, defined in names
        if any([s.islower() for s in a]):
            nuc = a[:2]
        else:
            nuc = a[0]
        # append to the all_nuc, which contains relevant nucleus
        all_nuc.append(nuc)

    # remove duplicates, to get a list of all unique parent elements included 
    all_nuc = sorted(set(all_nuc), key=lambda x: all_nuc.index(x))

    # the conservation of each included element
    # each row defines the names of the atomic species
    cons_out = []
    multi_out = [[1.,1.,1.] for i in range(len(all_nuc))]

    # Add the first two rows, which have special rules
    cons_out.append(['HII','HI', 'HII', 0])
    cons_out.append(['HeII', 'HeI','HeII', 0])

    multi_out[0][2] = 0.
    multi_out[1][2] = 0.


    # Add the rest of the inputs
    for i in range(2, len(all_nuc)):
        nuc = all_nuc[i]
        temp = ['III', 'I', 'II', 'III']
        temp = [nuc + t for t in temp]
        cons_out.append(temp)

    outp = []
    outp2 = []
    outp3 = []
    # # each of the molecules used
    for mol in all_mol:
    #    c is the charge (0 or 1)
        if 'II' in mol:
            charge = 1.
        else:
            charge = 0.

        # remove info about I or II or etc
        m = mol.replace('I', '')

        
        mlen = len(m)
    #    p1 and p2 are the parent, m1 and m2 are the multiplicity for parent 1 and 2 respectively
        p1 = 'x'
        p2 = 'x'
        m1 = 1.
        m2 = 1.
    #    finds the different special cases to 
    #    -------------------------------------
    #    if there is a digit, general 2 cases:
        if any([s.isdigit() for s in m]):
    #    1. X2 (double of same species)
            if mlen == 2:
                m1 = float(m[1])
                p1 = m[0]
    #    2. X2Y or XY2
            elif mlen > 2:
                digindi = np.where([s.isdigit() for s in m])[0][0]
    #    2.1 X2Y
                if digindi != mlen-1:
                    m1 = m[digindi]
                    p1 = m[:digindi]; p2 = m[digindi+1:]
    #    2.2 XY2
                else:
                    m2 = m[-1]
                    mtemp = m[:-1]
    #    2.21 XY2, with X and Y being only one letters long
                    if len(mtemp) == 2:
                        p1 = mtemp[0]; p2 = mtemp[1]
    #    2.22 XY2, however Xi or Yi
                    else:
                        lowindi =  [i for i, c in enumerate(m) if c.islower()]
                        if lowindi[0] != len(mtemp) - 1:
                            p1 = mtemp[:2]; p2 = mtemp[2:]
                        else:
                            p1 = mtemp[:1]; p2 = mtemp[1:]
    #    3. No digit, finds instances with XY, XiY and XYi
        else:
            if mlen == 2:
                p1 = m[0]; p2 = m[1]
            else:
                lowindi =  [i for i, c in enumerate(m) if c.islower()]
                if lowindi[0] != mlen-1:
                    p1 = m[:2]; p2 = m[2:]
                else:
                    p1 = m[:1]; p2 = m[1:]

    #    cleans up p1 and p2
        p1 = p1.strip(); p2 = p2.strip()
    #    cleans up multiplicity
        m1 = float(m1); m2 = float(m2)

        print(m, p1, p2, m1, m2, charge)

    #    prepears for nuccon output

        indi =  [i for i in range(len(cons_out)) if p1+'II' in cons_out[i]][0]
        cons_out[indi].append(mol)
        multi_out[indi].append(m1)
        if p2 != 'x':
            indi =  [i for i in range(len(cons_out)) if p2+'II' in cons_out[i]][0]
            cons_out[indi].append(mol)
            multi_out[indi].append(m2)

        # nuccon2.csv output stuff
        if p2 != 'x':
            outp2.append([return_bincode(p1 + 'I'),return_bincode(p2 + 'I'), m1, m2, return_bincode(mol), charge])
        else:
            outp2.append([return_bincode(p1 + 'I'),'0', m1, m2, return_bincode(mol), charge])
        
        outp3.append([return_bincode(mol), return_bincode(p1 + 'I')])
#    print(outp2)
    outp = []

    #makes the nuccon output format
    for i in range(len(cons_out)):
        cons = cons_out[i]
        temp = [0]*41

        for j in range(len(cons)):
            ctemp = cons[j]
            temp[j] = return_bincode(ctemp)

        for j in range(len(multi_out[i])):
            temp[21+j] = multi_out[i][j]

        outp.append(temp)

    #writes the nuccon output
    with open('nuccon_input.csv', mode='w') as out_file:
        out_write = csv.writer(out_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        out_write.writerow([len(outp), len(outp[0])])
        [out_write.writerow(o) for o in outp]

    #writes the nuccon2 output
    with open('nuccon_input2.csv', mode='w') as out_file:
        out_write = csv.writer(out_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        out_write.writerow([len(outp2), len(outp2[0])])
        [out_write.writerow(o) for o in outp2]
    
    with open('nuccon_input3.csv', mode='w') as out_file:
        out_write = csv.writer(out_file, delimiter=',', quotechar='"', quoting=csv.QUOTE_MINIMAL)
        out_write.writerow([len(outp3), len(outp3[0])])
        [out_write.writerow(o) for o in outp3]

suib_input_writer('../hr_umistnetwork_co.csv', 'mr_umistnetwork_co.csv')

_fcons()
