SUBROUTINE CREATEQ(T, atom, Nsize, Q, ne)	                ! Calculates collisional rates q from Omega values and temperature (Red Osterbrock)
USE M1, ONLY : atomm
USE PARAMETERS, ONLY : k
USE INTERFACE, ONLY : feI_collisionrates, feII_collisionrates, hI_coll_internal, hI_coll_differentn, mgI_collrates
IMPLICIT NONE
DOUBLE PRECISION, INTENT(IN) :: T
TYPE(atomm), INTENT(IN) :: atom
INTEGER, INTENT(IN) :: Nsize
DOUBLE PRECISION, DIMENSION(Nsize,Nsize), INTENT(OUT) :: Q
DOUBLE PRECISION :: ne
! =====================================================
! Task : Create collision rates Q
! INPUT : T         Temperature
!         atom      Type atomm (M1)
!         Nsize     Size of the collision matrix desired
!         ne        Electron number density [cm^-3]
! OUTPUT : Q        Collision rates [cm^3 s^-1]
! Globals :        
! =====================================================
DOUBLE PRECISION, PARAMETER :: const=8.629e-6	                        ! Osterbrock page 51
DOUBLE PRECISION, DIMENSION(Nsize,Nsize):: UPSILON	        ! Collision strengths
DOUBLE PRECISION :: dE, g1, g2, sumq1, sumq2
INTEGER :: i,j, NN, Nmax, NsizeH
DOUBLE PRECISION, DIMENSION(10,10) :: T1  ! Special for Fe I
DOUBLE PRECISION, DIMENSION(16,16) :: T2  ! Special for Fe II	                                                ! Loop variables
DOUBLE PRECISION, DIMENSION(11,11) :: T3 ! Special for MgI
DOUBLE PRECISION, DIMENSION(Nsize,Nsize) :: TEMPUSE, T0USE, BETA
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: QTEMP

! ---------------------------------------------------------

T1 = 0
T2 = 0
UPSILON = 0
Q = 0

! -------------------------------------------------------------

TEMPUSE = atom%UPSILON0(1:Nsize,1:Nsize)
T0USE = atom%UPSILON0_T0(1:Nsize,1:Nsize)
BETA = atom%UPSILON0_BETA(1:Nsize,1:Nsize)

WHERE (T0USE > 0)   ! If temperature-dependence exisits for the collision strengths. In readlevelsandlines, if there is no beta factor, T0 is set to zero so can use this criterium

   UPSILON = TEMPUSE*(T/T0USE)**BETA

ELSEWHERE

   UPSILON = TEMPUSE

END WHERE

IF (atom%NAME .EQ. 'FeI  ') THEN

   CALL feI_collisionrates(T, T1)  ! T1 contains collision strengths

   NN = MIN(Nsize,10)

   UPSILON(1:NN,1:NN) = T1(1:NN,1:NN)

ELSE IF (atom%name .EQ. 'FeII ') THEN

   CALL feII_collisionrates(T, T2)  ! T2 contains collision strengths

   NN = MIN(Nsize,16)

   UPSILON(1:NN,1:NN) = T2(1:NN,1:NN)

ELSE IF (atom%NAME .EQ. 'MgI  ') THEN

   call mgI_collrates(T,T3)

   NN = MIN(Nsize,11)

   do j = 2, NN
	do i = 1, j-1

	   if (T3(j,i) > 0.) then

   		UPSILON(j,i) = T3(j,i)

            end if	

         end do

    end do

END IF

DO j = 2, Nsize                                		! Lower levels

   DO i = 1, j-1			                        ! j is always larger than i
	
      dE = atom%LEVELS_E(j)-atom%LEVELS_E(i)		                ! Energy of line (always positive since j always larger)

      g1 = atom%levels_g(i)				                ! Get lower level degeneracy

      g2 = atom%levels_g(j)				                ! Get upper level degeneracy

      Q(j,i) = const*T**(-0.5)*UPSILON(j,i)/g2	                ! UPSILON Coefficients stored in lower left part of matrix

      Q(i,j) = Q(j,i)*g2/g1*EXP(-dE/(k*T))	                

   END DO

END DO

!printed the total excitation
!IF( atom%NAME .EQ. 'CMI  ') THEN
!    sumq1 = 0.
!    sumq2 = 0.
!    DO j = 67, 99                                		! Lower levels
!        DO i = 34, 66
!            sumq1 = sumq1 + Q(i,j)
!            sumq2 = sumq1 + Q(j,i)

!        END DO
!    END DO
!    print*,'printq',  Q(31, 1), Q(40, 10), Q(50, 20)
!    print*, 'de-excitation -> ', sumq2
!END IF

IF( atom%NAME .EQ. ' HI  ') THEN

!    Q = 0 ! CARE

    nmax = maxval([2,CEILING(-0.5 + SQRT(0.25 + 2*Nsize))])   ! collision routines can only be called with full principal quantum states.
    NsizeH = ((nmax-1)**2 + (nmax-1))/2 + nmax    ! The fine-structure-split size that would fill the current p.q.n.

    ALLOCATE(QTEMP(NsizeH,NsizeH)); QTEMP = 0 

    !l-changing transitions (same n) (different ones than in hI_coll_differentn)
    CALL hI_coll_internal(1, T, ne, NsizeH, DBLE(atom%A(1:NsizeH,1:NsizeH)), QTEMP)  

    DO j = 2, Nsize

       DO i = 1, Nsize

          if (QTEMP(j,i) .ne. 0) then

             Q(i,j) = QTEMP(i,j)
             Q(j,i) = QTEMP(j,i)

          end if

       END DO

    END DO

    ! n-changing transitions (different ones than in hI_coll_internal):
    CALL hI_coll_differentn(T, NsizeH, atom%LEVELS_E(1:NsizeH), atom%LEVELS_G(1:NsizeH), DBLE(atom%A(1:NsizeH,1:NsizeH)), QTEMP)

    DO j = 22, Nsize ! First five n (i=1-21) covered by Anderson2000

       DO i = 1, Nsize

          IF (QTEMP(j,i) .ne. 0) then

             Q(i,j) = QTEMP(i,j)
             Q(j,i) = QTEMP(j,i)

           END IF

       END DO

    END DO

END IF

END
