SUBROUTINE createbetaandQvec(atom, from, Nup, EXC_MAT, nmolvec, ndensvec_abin, alfa_dust, N_atoms, atoms, R_index, T, ne, beta_s, beta_c, Qvecdown, Qvecup)
USE PARAMETERS, ONLY : k, mp, pi
USE M1, ONLY : atomm, time, Nabin, Npis, Nmol, Nelem, ngvec, Nstages
USE INTERFACE, ONLY : beta_s_func, feI_collisionrates, feII_collisionrates, alfa_photoionization, hI_coll_internal, hI_coll_differentn
IMPLICIT NONE
TYPE(atomm), INTENT(IN) :: atom
INTEGER, INTENT(IN) :: from, R_index, Nup, N_atoms
REAL, DIMENSION(:,:), INTENT(IN) :: EXC_MAT
DOUBLE PRECISION, DIMENSION(maxval([1,Nmol])), INTENT(IN) :: nmolvec
DOUBLE PRECISION, DIMENSION(Nabin), INTENT(INOUT) :: ndensvec_abin
TYPE(atomm), DIMENSION(N_atoms), INTENT(IN) :: atoms
DOUBLE PRECISION, INTENT(IN) :: T, ne, alfa_dust
DOUBLE PRECISION, DIMENSION(from-1), INTENT(OUT) :: beta_s, beta_c, Qvecdown
DOUBLE PRECISION, DIMENSION(Nup), INTENT(OUT) :: Qvecup
! ==========================================================
! TASK :  Computes the escape probabilities, cont. destruction probabilities and collisional deexcitation/excitation probabilities for an atom in a particular excited state.
! INPUT : atom            Struct type atomm (M1)
!         from            Index of the excited state the atom is in
!         Nup             Number of upwards collisional transitions to include
!         EXC_MAT         Population solutions [cm^-3]
!         nmnolvec        Vector with molecular densities [cm^-3]
!         ndensvec_abin   Array with atomic number densitues [cm^-3]
!         alfa_dust       Dust opacity [cm^-1]
!         N_atom          Number of atoms
!         atoms           array with all atoms (struct type atomm)
!         R_index         Current zone
!         T               Temperature
!         ne              Electron density [cm^-3]
! OUTPUT: beta_s          Sobolev escape probability
!         beta_c          Continuum destruction probability
!         Qvecdown        Collisional deexcitation rates
!         Qvecup          Collisional excitation rates
! ==========================================================
DOUBLE PRECISION, PARAMETER :: const=8.629D-6
INTEGER :: i, Nsize, iuse, nmax
DOUBLE PRECISION :: g1, g2, tau, dnu, alfa_c, alfa_l, psi, f, lambda, lambda_c_cut, aconst, beta_c_temp
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: Q
DOUBLE PRECISION, DIMENSION(10,10) :: T1
DOUBLE PRECISION, DIMENSION(16,16) :: T2
DOUBLE PRECISION, DIMENSION(from-1+Nup) :: upsilonvec
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: ALFA_MAT
!DOUBLE PRECISION, EXTERNAL :: beta_s_func
! -------------------------------------------------------------------

ALLOCATE(ALFA_MAT((Nstages-1)*Nelem+Nmol+3,maxval([Npis,ngvec]))); ALFA_MAT = 0

Nsize = minval([from + Nup,atom%N])

ALLOCATE(Q(Nsize,Nsize)); Q = 0

lambda_c_cut = 5000.   ! CARE photons above this wavelength automatically get beta_c = 0 (the process is mostly important in the UV)

upsilonvec = 0

beta_s = 0

beta_c = 0

Qvecdown = 0

Qvecup = 0

! ---------------------------------------------

g2 = atom%levels_g(from)    ! Statistical weight of upper level

call createQ(T, atom, Nsize, Q, ne)

! Radiative deexcitations:

DO i = 1,(from-1)   ! To all states below..

   g1 = atom%LEVELS_G(i)

   IF (i <= atom%N_LEVELS_ABSORB) THEN  ! CARE can fail if if n_levels_absorb is screwed up

      tau = atom%TAU(R_index,i,from)

   ELSE

      tau = 0

   END IF

   beta_s(i) = beta_s_func(tau)

   Qvecdown(i) = Q(from,i)

   lambda = atom%LAMBDA(from,i)

   ! Continuum destruction probability:

   IF (lambda < lambda_c_cut .AND. tau > 10.) THEN ! CARE..only bother is tau over some threshold otherwise continuum destruction unimportant

      CALL alfa_photoionization(lambda, N_atoms, atoms, EXC_MAT, nmolvec, alfa_dust, T, ne, ndensvec_abin, alfa_c, ALFA_MAT) ! Get the continuum absorption coefficients

      dnu = SQRT(2*k*T/(atom%MASS*mp))/(lambda*1E-8)  ! Thermal line width

      alfa_l = tau/(lambda*1E-8*dnu*time)   ! See Hummer & Rybicki 1985

      psi = 1./tau

      aconst = sum(atom%A(from,1:from-1))/(4*pi*dnu)

      call createbeta_c_single(tau, alfa_c/alfa_l, aconst, beta_c_temp)

      beta_c(i) = beta_c_temp

   END IF

END DO

! Upward collisions:

DO i = 1, Nup

   IF (from <= atom%N_LEVELS .AND. from + i <= atom%N_LEVELS) Qvecup(i) = Q(from,from+i)

END DO

IF (Nup == 1) Qvecup = 0 ! CARE no upward collisions allowed in fluorescence if Nup equals 1 (it cannot be set to zero)

END
