SUBROUTINE excitation_solver(name, N_levels, N, AMAT, QMAT, BETAMAT, NTEXCMAT, LYMAT, gammaion, GAMMAPHOTOEXC, alphain, threebodyrates, alphaout, ctanddirection, nk, nplus, ne, levels_pop, FAIL, NEG)
USE INTERFACE, ONLY : DGESV
USE M1, ONLY : PHOTOEXC, globid, IDP, DOPRINT
IMPLICIT NONE
CHARACTER(len=5) :: name
INTEGER, INTENT(IN) :: N_levels, N
DOUBLE PRECISION, DIMENSION(N_levels), INTENT(IN) :: alphain, threebodyrates, gammaion, ctanddirection
DOUBLE PRECISION, DIMENSION(N_levels, N_levels), INTENT(IN)  :: AMAT, QMAT, BETAMAT, NTEXCMAT, LYMAT
DOUBLE PRECISION, DIMENSION(N_levels, N), INTENT(IN) :: GAMMAPHOTOEXC
DOUBLE PRECISION, INTENT(IN) :: nk, nplus, ne, alphaout
DOUBLE PRECISION, DIMENSION(N_levels), INTENT(OUT) :: levels_pop
INTEGER, INTENT(OUT) :: FAIL, NEG
! =================================================================================
! TASK : Solves atomic level populations for one atom by computing statistical equilibrium
! INPUT : N_LEVELS : Number of levels in atom to solve for
!         N          Total number of levels in atom
!         AMAT : A values (upper level is first index, lower second) [s^-1]
!         QMAT : Collision rates (ne not included) [cm^3 s^-1].
!         BETAMAT : The loss probability (chance of photon not being re-absorbed in same line) [dim.less]. First index is upper level.
!         NTEXCMAT : Nonthermal excitation rates [s^-1]. First index is lower level.
!         LYMAT    : Pumping by Ly radiation. [s^-1]. First idndex is lower level.
!         gammaion    Ionization rates  [s^-1]
!         GAMMAPHOTOEXC  Photoexcitation rates (destruction rates, assume ground state is target (via cascade which is emitted in MC)) [s^-1]
!         alphain     Specific Recombination rates [cm^3 s^-1]
!         alphaout    Specific rec rates out of atom [cm^3 s^-1]
!         ctanddirection        Charge transfer rates to excited levels [s^-1]
!         nk          Number density of atom [cm^-3]
!         nplus       Number density of ion above atom [cm^-3]
!         ne          Electron number density [cm^-3]
! OUTPUT :levels_pop : Solution for level populations [cm^-3]
!         FAIL        Boolean, did solution fail?
!         NEG         Boolean, were there negative population solutions?
! GLOBALS :
! ==============================================================================
DOUBLE PRECISION, DIMENSION(N_levels,N_levels) :: MMAT, CASCMAT, MMATDUMMY			        ! Matrix for linear equation system to be created
DOUBLE PRECISION :: summ, rectot
INTEGER :: i,j, N_levels_print, info, icheck, Ncheck		                                                        ! Indeces
INTEGER, DIMENSION(N_levels) :: ipiv	                                        ! SGESV output signal
DOUBLE PRECISION, DIMENSION(N_levels) :: bdummy
DOUBLE PRECISION :: photoexctot, collexctot, ntexctot, directrec, fromabove, ttt

! --------------------------------------------------

Ncheck = MIN(25, N_levels)

FAIL = 0
NEG = 0                                    ! Important

! ------------------------------------------------------------

DO i = 1, N_levels

   summ = 0						! have to reset sum for every eye index (i,i)

   DO j = 1, (i-1)   ! INTERACTIONS WITH LOWER LEVELS (j < i)

      MMAT(i,j) = QMAT(j,i)*ne + NTEXCMAT(j,i) + LYMAT(j,i)	             ! Gains from lower levels: collisional excitations, non-thermal excitations and possibly Ly-alpha pumping

      IF (PHOTOEXC == 2) THEN

         MMAT(i,j) = MMAT(i,j) + GAMMAPHOTOEXC(j,i)  ! CARE    ! Fill upper level from photoexcitations, j < i

      END IF

      summ = summ + AMAT(i,j)*BETAMAT(i,j) + QMAT(i,j)*ne	 ! Losses to lower levels: spontanous deexcitation (ignore stimulated) and collisional deexcitation

   END DO

   DO j = (i+1), N_levels        ! INTERACTIONS WITH UPPER LEVELS (j > i)

      MMAT(i,j) = AMAT(j,i)*BETAMAT(j,i) + QMAT(j,i)*ne	        ! Gains from higher levels: spontanous deexcitation and collisional deexcitation

      IF (PHOTOEXC == 2) THEN

          MMAT(i,j) = MMAT(i,j) + GAMMAPHOTOEXC(j,i) ! CARE  ! gains from Stimulated emission

      END IF

      summ = summ + QMAT(i,j)*ne + NTEXCMAT(i,j) + LYMAT(i,j)	       ! Losses to higher levels: collisional excitation, non-thermal excitations and possibly Ly-alpha

   END DO

   MMAT(i,i) = -summ - gammaion(i)   ! Add photoionizations

   MMAT(i,i) = MMAT(i,i) - alphaout*ne   ! Recombinations out (to the ion below)

   IF (PHOTOEXC == 2) THEN

      MMAT(i,i) = MMAT(i,i) - sum(GAMMAPHOTOEXC(i,1:N_levels))	                ! Add losses to all other levels by stimulated emission

   ELSEIF (PHOTOEXC == 1) THEN

      MMAT(i,i) = MMAT(i,i) - sum(GAMMAPHOTOEXC(i,1:N))

   END IF

   ! PHOTOEXC = 0 : No gamma rates at all (no energy is created/destroyed in transfer)

   bdummy(i) = -alphain(i)*ne*nplus - ctanddirection(i) -threebodyrates(i)*ne**2*nplus                       ! Level filling by recombinations...CARE should add CT-reactions to excited states here

END DO

MMAT(1,:) = 1.                   ! Replace one equation with density constraint

bdummy(1) = nk                   ! Sum of all levels should be total el abundance

MMATDUMMY = MMAT  ! Use dummy in DGESV as it modifies matrix

N_levels_print = MIN(N_levels, 35)

IF (globid .EQ. IDP .AND. DOPRINT) THEN
!   print*, QMAT(42,1), QMAT(1,42)
!   PRINT*, QMAT(139,120), QMAT(120, 139)
   print *, 'N_levels_print', N_levels_print

   !PRINT *, 'gammaion:  gammaphotoexc:'
   !DO i = 1, N_levels_print
   !   PRINT *, i, gammaion(i), sum(gammaphotoexc(i,1:N_levels))
   !END DO
   PRINT *, 'A: '
   CALL matrixer(N_levels_print, AMAT(1:N_levels_print,1:N_levels_print))
   !PRINT *, 'BETAMAT: '
   !CALL matrixer(N_levels_print, BETAMAT(1:N_levels_print,1:N_levels_print))
   PRINT *, 'Q*ne: '
   CALL matrixer(N_levels_print, QMAT(1:N_levels_print,1:N_levels_print)*ne)
   !PRINT *, 'NTEXCMAT: '
   !CALL matrixer(N_levels_print, NTEXCMAT(1:N_levels_print,1:N_levels_print))
   !PRINT *, 'GAMMAEXCMAT: '
   !CALL matrixer(N_levels_print, GAMMAPHOTOEXC(1:N_levels_print,1:N_levels_print))
   !PRINT *, 'LYMAT: '
   !CALL matrixer(N_levels_print, LYMAT(1:N_levels_print,1:N_levels_print))
   PRINT *, 'M: '
   CALL matrixer(N_levels_print, MMAT(1:N_levels_print,1:N_levels_print))
   !PRINT *, 'Rec:   Coll rec:   CT+D.I.'
   !DO i=1, N_levels_print
   !     write(*,'(3ES12.2)'),  alphain(i)*ne*nplus, threebodyrates(i)*ne**2*nplus, ctanddirection(i)
   !END DO
   !print *, 'alphaout', alphaout*ne

   !print *, 'Rectot', sum(alphain), sum(alphain)*ne*nplus

END IF

! Call NR routine for solving a linear equation system:

CALL DGESV(N_levels, 1, MMATDUMMY, N_levels, ipiv, bdummy, N_levels, info)

IF (info .NE. 0) THEN
   WRITE(*,*), 'WARNING FAILURE in DGESV (in excitation_solver).'
   FAIL = 1    ! Dont stop execution here but return a fail signal to calling routine
END IF

levels_pop = bdummy

WHERE (levels_pop < 0)  ! Dont allow negative populations

   levels_pop = 0  ! CARE

END WHERE

if (levels_pop(1) == 0) then

   levels_pop(1) = nk*0.9 ! TEMP

   print *, 'WARNING, neg.gs pop.i=1,shell ', globid, name

   NEG = 1

end if

!call cascadematrix(N_levels, AMAT*BETAMAT, CASCMAT)  !WARNING this takes very long time. CASC(j,i) gives the fraction of populations of j that eventually reach i

IF (DOPRINT .AND. globid == IDP) THEN

   WRITE(*,*) 'Sol: ', levels_pop(1:N_levels_print)

   !photoexctot = sum(GAMMAPHOTOEXC(1,2:4))*levels_pop(1)! TOtal number of photoexc into ICHECK

   !collexctot = sum(QMAT(1,2:4))*ne*levels_pop(1) ! Total number of thermal collisional excitations

   !ntexctot =  sum(NTEXCMAT(1,2:4))*levels_pop(1)

   !directrec = ne*nplus*sum(alphain(2:4))

   !fromabove =  sum(MMAT(2,5:N_levels)*levels_pop(5:N_levels)) + sum(MMAT(3,5:N_levels)*levels_pop(5:N_levels)) + sum(MMAT(4,5:N_levels)*levels_pop(5:N_levels))

   !ttt = photoexctot + collexctot + ntexctot + directrec + fromabove

   !print *, '4571apa photoexc, collexc, ntexc, directrec, fromabove:', photoexctot/ttt, collexctot/ttt, ntexctot/ttt, directrec/ttt, fromabove/ttt, sum(MMAT(2:4,6))*levels_pop(6)/ttt

   !rectot = sum(alphain(icheck:N_levels)*CASCMAT(icheck:N_levels,icheck))  ! This should be the effective rec coefficient

   DO icheck = 1, Ncheck

      photoexctot = sum(GAMMAPHOTOEXC(1:icheck-1,icheck)*levels_pop(1:icheck-1))! TOtal number of photoexc into ICHECK

      collexctot = sum(QMAT(1:icheck-1,icheck)*ne*levels_pop(1:icheck-1)) ! Total number of thermal collisional excitations

      ntexctot =  sum(NTEXCMAT(1:icheck-1,icheck)*levels_pop(1:icheck-1))

      directrec = ne*nplus*alphain(icheck)

      fromabove =  sum(MMAT(icheck,icheck+1:N_levels)*levels_pop(icheck+1:N_levels))

      ttt = photoexctot + collexctot + ntexctot + directrec + fromabove + 1D-50

      print *, icheck, 'superapa photoexc, collexc, ntexc, directrec, fromabove:', photoexctot/ttt, collexctot/ttt, ntexctot/ttt, directrec/ttt, fromabove/ttt

      !rectot = sum(alphain(icheck:N_levels)*CASCMAT(icheck:N_levels,icheck))  ! This should be the effective rec coefficient

      IF (icheck <= N_levels .AND. ne*nplus > 0) THEN

	PRINT *, 'level', icheck
        print *, 'inflow via levels above (radiative)', SUM(AMAT(icheck+1:N_levels,icheck)*BETAMAT(icheck+1:N_levels,icheck)*levels_pop(icheck+1:N_levels))
       print *, 'this rate relative to total recombination flow', SUM(AMAT(icheck+1:N_levels,icheck)*BETAMAT(icheck+1:N_levels,icheck)*levels_pop(icheck+1:N_levels))/(SUM(alphain)*ne*nplus)
        print *, 'direct rec flow', alphain(icheck)*ne*nplus, alphain(icheck)

      end if

   END DO

END IF

20 FORMAT(A10,100ES10.2)


END
