SUBROUTINE hI_coll_differentn(T, N_levels, Evec, gvec, A, Q)
USE PARAMETERS, ONLY : k
IMPLICIT NONE
DOUBLE PRECISION, INTENT(IN) :: T
INTEGER, INTENT(IN) :: N_levels
DOUBLE PRECISION, DIMENSION(N_levels), INTENT(IN) :: Evec, gvec   ! Enery levels (erg), stat weights
DOUBLE PRECISION, DIMENSION(N_levels, N_levels), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(N_levels, N_levels), INTENT(INOUT) :: Q     ! CARE dont put Q=0, Q comes in with some value already that should not be overwritten
! =================================================================
!  This routine calculates the collisional excitation rate (cnl) 
!  between different n-states for h i
!
!  Hummer & Storey, 1987 (MNRAS, 224, p. 801) uses the Bethe-approx.,
!  Q_nl,n'l' = (f_nl,n'l'/f_nn')*Q_nn',
!  for transitions with delta(n)=1. For delta(n) > 1 they use a simpler   AJ care Q is a cross section here, a cross section maps to Upsilon/g
!  approximation, Q_nl,n'l' = Q_nn'. Only transitions with delta(l)=1     AJ Hummer Storey write that using Q_nl,nl' = Q_nn' gives very similar results as doing as abov!                                                                         e, the specific l-dependence on n-changing collisions are not important.
!  are considered.
!   
!  For all transitions (delta(n) > 0, delta(l)=1), we use the 
!  Bethe-approximation. Our treatment differs from that by Hummer & Storey
!  in that we aslo include the statistical weights,
!  C_nl,n'l' = (g_n/g_nl)*(f_nl,n'l'/f_nn')*C_nn' (g_n/g_nl = 2n^2/2(2l+1)).
!  This assumption  gives :
!     x_n * C_nn' = sum(ll'; x_nl * C_nl,n'l') if the different l-states
!  are assumed to be populated according to statistical weights.
!
!  The collisional excitation rates (C_nn') are from johnson (ap j 174:227).
! =====================================================================

INTEGER :: nmax, i, counter, n, poslong, n_lower, n_upper, lower_pos, upper_pos, j, l_down, l_up
DOUBLE PRECISION :: cs12s, cs12p, et, C_up, C_down, Atot, gi, gj, fact, Tuse, et_lim, term, const
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: Eprincipal, gprincipal

! ----------------------------------------------------------

et_lim = 30.       ! Limiting Boltzmann factor

! _________________________________________

Q = 0

nmax = NINT(-0.5 + SQRT(0.25 + 2*N_levels))  ! Ex N_levels = 3 --> n = 2, N_levels = 4 --> n=3

ALLOCATE(Eprincipal(nmax)); Eprincipal = 0

ALLOCATE(gprincipal(nmax)); gprincipal = 0  

DO n = 1, nmax   ! for each principal q.n.

   poslong = ((n-1)**2 + n + 1)/2   ! Starting pos for each principal n in nl array

   Eprincipal(n) = Evec(poslong)

   gprincipal(n) = SUM(gvec(poslong:poslong+n-1))  ! There are n sublevels for each principal n

END DO

!  Transition rates for 1-2s, 1-2p, and 1-3 from aggarwall mn 202, 15p '83. for 2s

cs12s = 0.195 + 1.28e-5*T - 4.69e-10*T**2 + 6.19e-15*T**3           ! Valid up to 25,000 K  0.25 at 5000 K
        
cs12p = 0.335 + 1.45e-5*T + 1.39e-10*T**2 - 5.66e-15*T**3      ! 0.41 at 5000 K
   
cs12p = cs12p*0.88   !  scale factor to agree with callaway unnikrishnan and oza

Q(2,1) = 8.629e-6/SQRT(T)*cs12s/2.   ! Osterbrock eq 3.18

Q(3,1) = 8.629e-6/SQRT(T)*cs12p/6.

Q(1,2) = Q(2,1)*2./2.*EXP(-1.1835E5/T)  ! osterbrock eq 3.20

Q(1,3) = Q(3,1)*6./2.*EXP(-1.1835E5/T)
      
DO n_lower = 1, nmax-1    ! for each lower principal q.n.
        
   DO n_upper = n_lower+1, nmax   ! for each upper principal q.n.

      IF (n_lower .EQ. 1 .AND. n_upper .EQ. 2) GOTO 9  ! This case (n=1 and n=2) is already treated 

      lower_pos = ((n_lower-1)**2 + n_lower + 1)/2   ! Startpos of lower principal n

      upper_pos = ((n_upper-1)**2 + n_upper + 1)/2   ! Startpos of upper principal n
      
      et = (Eprincipal(n_upper)-Eprincipal(n_lower))/(k*T)  ! Boltzmann exponent

      IF (et < et_lim) THEN ! dont use too large et since the boltzman factor in CEXH then becomes zero
         
         CALL CEXH(n_lower, n_upper, T, nmax, Eprincipal, C_up)  ! CEXH returns upward rates C(n_down --> n_up)

         IF (C_up < 0) C_up = abs(C_up) ! CARE 'cexh', n_lower, n_upper, C_up, T, Eprincipal(n_lower), Eprincipal(n_upper), N_levels

         IF (C_up .EQ. 0.) THEN; PRINT *, 'ERROR in hI_coll_differentn';stop 1;END IF
 
         C_down = C_up*gprincipal(n_lower)/gprincipal(n_upper)*dexp(et)  ! Downward rate  C(n_up --> n_down)

         IF (C_down < 0) print *, 'down', C_down
          
      ELSE   ! very low T --> upward boltzman factors risks becoming zero

         Tuse = (Eprincipal(n_upper)-Eprincipal(n_lower))/(k*et_lim)  ! the temperature that corresponds to et_lim

         CALL CEXH(n_lower, n_upper, Tuse, nmax, Eprincipal, C_up)

         IF (C_up < 0) C_up = abs(C_up)  ! CARE print *, 'cexh2', n_lower, n_upper, C_up
         
         !et = (Eprincipal(n_upper)-Eprincipal(n_lower))/(k*Tuse) 

         C_down = C_up*gprincipal(n_lower)/gprincipal(n_upper)*dexp(et_lim)*(T/Tuse)**(-0.5)  ! Real C_down, know scaling is T^-0.5

         IF (C_down < 0) print *, 'down2',n_lower, n_upper, C_down

      END IF

      term = 0  ! DO NOT REMOVE

      DO l_down = 0, n_lower-1  ! For each l state in the lower n...

         DO l_up = 0, n_upper-1  ! for each l state in the upper n..

            Q(upper_pos+l_up,lower_pos+l_down) = C_down*gvec(lower_pos + l_down)/gprincipal(n_lower)  ! Q(u,l) = Q(nu,nl)*glow/glowtot

            Q(lower_pos+l_down,upper_pos+l_up) = Q(upper_pos+l_up,lower_pos+l_down)*gvec(upper_pos + l_up)/gvec(lower_pos + l_down)*dexp(-et)  ! Q(l,u) = Q(u,l)*gu/gl*exp(-et)
                  
            IF (Q(upper_pos+l_up,lower_pos+l_down) < 0) print *, 'Q', upper_pos+l_up, lower_pos+l_down, Q(upper_pos+l_up,lower_pos+l_down)
            
            IF ((l_up .NE. l_down-1) .AND. (l_up .NE. l_down+1)) THEN  ! If not an l-changing transition, damp by 0.1

                ! FOR NOW, IGNORE effects of changing l or not its the only treatment that converges for some reason

                !Q(upper_pos+l_up,lower_pos+l_down) = 0.1*Q(upper_pos+l_up,lower_pos+l_down)

                !Q(lower_pos+l_down,upper_pos+l_up) = 0.1*Q(lower_pos+l_down,upper_pos+l_up) 

                !Q(upper_pos+l_up,lower_pos+l_down) = 0 ! 

                !Q(lower_pos+l_down,upper_pos+l_up) = 0 ! 

               continue  

            END IF
                 
            !gj = 2.*(2.*j+1)
             
            !fact = A(upper_pos+j, lower_pos+i)/Atot

            !Q(upper_pos+l_up,lower_pos+l_down) = const*gvec(lower_pos + l_down)*A(upper_pos + l_up, lower_pos + l_down)  
             
            !Q(upper_pos+l_up,lower_pos+l_down) = C_down*gvec(lower_pos + l_down)/gprincipal(n_lower)    ! Q(u,l) = Qdown*glow(i)/glowtot

            !Q(lower_pos+l_down,upper_pos+l_up) = Q(upper_pos+l_up,lower_pos+l_down)*gvec(upper_pos + l_up)/gvec(lower_pos + l_down)*dexp(-et)
            
         END DO
                
         !l_up = l_down - 1        ! and l-1 in the upper n.. (consider only l-changing collisions)
         
         !IF (l_up >= 0) THEN   ! check that l_down was not zero, then l_down-1 doesnt exist
    
            !gj = 2.*(2.*j+1)
             
            !fact = a(upper_pos+j,lower_pos+i)/Atot
             
            !Q(upper_pos+l_up,lower_pos+l_down) = const*gvec(lower_pos + l_down)*A(upper_pos + l_up, lower_pos + l_down)

            !Q(upper_pos+l_up,lower_pos+l_down) = C_down*gvec(lower_pos + l_down)/gprincipal(n_lower)  ! Q(u,l) = Qdown*glow(i)/glowtot

            !Q(lower_pos+l_down,upper_pos+l_up) = Q(upper_pos+l_up,lower_pos+l_down)*gvec(upper_pos + l_up)/gvec(lower_pos + l_down)*dexp(-et)
          
         !END IF

      ENDDO

9     CONTINUE
        
   ENDDO
      
ENDDO

END SUBROUTINE hI_coll_differentn

