SUBROUTINE hI_coll_internal(partner, T, ne, N_levels, A, Q)
USE PARAMETERS, ONLY : me, mp
USE M1, only: myid
IMPLICIT NONE
INTEGER, INTENT(IN) :: partner, N_levels
DOUBLE PRECISION, INTENT(IN) :: T, ne
DOUBLE PRECISION, DIMENSION(N_levels,N_levels), INTENT(IN) :: A
DOUBLE PRECISION, DIMENSION(N_levels,N_levels), INTENT(INOUT) :: Q                ! NOTE DONT PUT Q = 0 as initialization
! ===============================================
! Task : Computes delta-n=0 collision rates for HI.
! Input : partner           INTEGER, electrons=1, protons=2
!          T                Temperature
!          ne               Electron number density (cm^-3)
!          N_levels         Number of l states
!          A                N_levels*N_levels matrix with A values A(up,low)
! Output : Q                Collision rates (cm^3/s)         
! =============================================== 
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: anl, c1p, c1m, c2p, c2m 
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: cl    
DOUBLE PRECISION :: my, mym, C_nl1, vel, rho_D, two_log_rho_D, tau, rho_c, D_nl, C_nl, two_log_rho, cnl1, two_log_rho_1, two_log_rho_2
INTEGER :: l, nl, np, lp, nplp, nmax, n, istart, pos1
     
!  CK comments :  References :
!     Brocklehurst, M., 1971, MNRAS, 153, P. 471-490. (B71)
!     Pengelly, R.M.,& Seaton, M.J., 1964, MNRAS, 127, p. 165. (PS64)
!
!  The DATA is checked by reproducing figure 2. in PS64, in detail.
!  That is to plot xltc (in main code) as a FUNCTION of l, for 
!  collsions WITH protons, T=1.e4K, and Ne=1.e4.
! 
!  Points of confusion :
!
!  - Both PS64 (eq. 33) and B71 (eq. 3.36) uses an expression for
!    the Debye radius containg a 'pi' too much. B71 THEN uses this
!    'wrong' Debye radius in the following expressions (eq. 3.40).   
!    PS64, on the other hand uses the correct Debye radius (without
!    the extra pi) in the corresponding expression (eq. 46). In PS64 
!    it mostly looks like a misprint in eq. 33, when calculating the 
!    expression they get the correct number.
!    In my calculation I USE the correct Debye radius in all expressions. 
!
!  - I can't derive eq. 3.41 in B71 ( and the corresponding eq. 45 in
!    PS64). They get 2log10(0.72*v*tau)=10.95+log10(m*T*tau*tau/my).
!    By using the following expression for the velocity,
!    v = sqrt(2kT/my) (this is the same expression as given by PS64
!    in eq. 31, however when calculating it explicitly they get the
!    wrong answer), I get the expression
!    2log10(0.72*v*tau)=11.196+log10(m*T*tau*tau/my).
!    I'm using this last expression.
!
! zZ = z/Z ; WHERE z is the charge of the particle causing the excitation,
!            (z=1 for electrons), Z is the charge of the nuclear charge 
!            (Z=1 for hydrogen). 
!
!
! collisions WITH electrons : partner=1
! collisions WITH protons : partner=2
!
!  mym = my/m = reduced mass / electron mass

Q = 0

nmax = NINT(-0.5 + SQRT(0.25 + 2*N_levels))   ! Number of principal q. numbers corresponding to N_levels nl states

DO n = 2, nmax    ! For each principal q.n.

   ALLOCATE(anl(0:n*(n+1)/2)); anl = 0
   ALLOCATE(c1p(0:n)); c1p = 0
   ALLOCATE(c1m(0:n)); c1m = 0
   ALLOCATE(c2p(0:n)); c2p = 0
   ALLOCATE(c2m(0:n)); c2m = 0
   ALLOCATE(cl(0:n,0:n)); cl = 0
   
   IF (partner .EQ. 1) THEN   ! collisions w electrons

      !  my = mp*me/(mp+me) = me 
      !  mym = my / me = 1.
         
      my = me    ! reduced mass is (m1*m2/(m1+m2) ~ mH*me/mH = me
         
      mym = 1.   ! mym = reduced mass / electron mass = me / me = 1
      
   ELSEIF (partner .EQ. 2) THEN  ! collisions w protons

      !  my = mp*mp/(mp+mp) = 0.5 mp 
      !  mym = my / me = 0.5 mp/me
                  
      my = 0.5*mp   ! reduced mass = mp^2/(2mp) = 0.5mp
         
      mym = my/me   ! mym = reduced mass / electron mass 
      
   ENDIF

   Cnl1 = 9.933d-6*SQRT(mym)/SQRT(T)    ! PS64 eq 43  first part
      
   vel = SQRT(2.*1.3806d-16*T/my)                  ! PS64 eq 31  cm/sec  note that reduced mass is used and not particle mass..not sure why

   !  rho_c = 0.72*tau*v ; tau = radiative lifetime; v = velocity of 
   !                       the incident particle
   !  (rho_D = SQRT(kT/4*pi^2*Ne*e^2) = Debye radius/SQRT(pi) (?); not used !)
   !  rho_D = SQRT(kT/4*pi*Ne*e^2) = Debye radius  ; used !
   !  tau = radiative lifetime

   !      rho_D=SQRT(47.68*te/ne)  correct -->
   !      two_log_rho_D=1.68+LOG10(te/ne)   

   rho_D = SQRT(47.6248*T/ne)                    !Debye radius    BH 3.36 one pi removed (typo se above)  PS eq 33 (they have one pi too much)
                       
   anl = 0

   istart = ((n-1)**2 + (n+1))/2   ! starting nl value for each n CHECKED

   DO l = 0, n-1 ! For a given n, there are n different l values
         
      anl(l) = SUM(A(istart+l,1:istart))    ! Add up all A values out from state nl to all lower states

   END DO

   ! start from l=0 and climb up (c1p & c1m)
      
   l = 0

   if (anl(l) > 0) then
            
   	tau = 1./anl(l)       ! lifetime if this nl state is is 1/A

   elseif (n == 2) then

      	tau = 1/8.29 ! Two photon decay

   else
	
	tau = 1e5 ! CARE made up value
        print *, "WARNING', anl(l) = 0 condition hit in hI_coll_internal", n

   end if

   rho_c = 0.72*vel*tau    ! PS eq 30
     
   two_log_rho_1 = 1.68 + LOG10(T/ne)     ! PS eq 46   adjusted by CK (see above)   If rho_D is smallest

   two_log_rho_2 = 11.196 + LOG10(me*T*tau**2/my)   ! PS eq 45     If rho_c is smallest

   two_log_rho = MINVAL([two_log_rho_1, two_log_rho_2])  ! PS64 page 170 "of the three expressions Rc1, Rc2, Rd, we adopt whichever is smallest". Here, we ignore calculations of Rc1 = 1.12 hnu/deltaE (which is usually similar to 0.72v*tau). This is done also in Brocklehurst1971

   D_nl = 6*n**2*(n**2 - l**2 - l - 1.)   ! PS64 eq 44  Z/z = 1 for electrons and protons (but would be 2 for HeII ions)
      
   C_nl = cnl1*D_nl*(11.538 + LOG10(T/(D_nl*mym)) + two_log_rho)   ! PS64 eq 43

   c1p(0) = C_nl  ! Brocklehurst1971 eq 3.42   c1p is allocated to start at index 0
            
   c1m(1) =((2.*l+1.)/(2.*(l+1.)+1.))*c1p(0)   ! m arrays contain downward rates. Detailed balance, C_nl,nl+1 * g(l) =  = C_nl+1,nl * g(l+1) --> C_nl+1,nl = g(l)/g(l+1)*C_nl,nl+1

   DO l = 1, n-2   ! For all l values except l=0 (treated above) also skip l = n-1 because then l_lower = lmax no upwards transition can occur
         
      tau = 1./anl(l)   ! new A -> new lifetime
         
      two_log_rho_2 =  11.196 + LOG10(me*T*tau**2/my)  ! new tau, recompute rho_c
         
      two_log_rho = MINVAL([two_log_rho_1,two_log_rho_2])   ! rho_D does not depend on l, dont recompute
         
      D_nl = 6*n**2*(n**2- l**2 - l -1.)
         
      C_nl = cnl1*D_nl*(11.538 + LOG10(T/(D_nl*mym)) + two_log_rho)  ! This is C_n1,n2 + C_n1,n0 (for l = 1)

      c1p(l) = C_nl - c1m(l)  ! up rate C_nl,n2 = C_n1 - C_n1,n0 (see line above) = C_n1 - c1m(1)   
                  
      c1m(l+1) = ((2.*l+1.)/(2.*(l+1)+1.))*c1p(l)  ! Detailed balance C(l=2 to l=2) = g(l=1)/g(l=2)*C(l=1 to l=2)
      
   ENDDO

   ! We can also start from l=n-1 and climb down (c2p & c2m) See Brocklehurst1971

   l = n-1    ! highest l value 
       
   tau = 1./anl(l)  ! lifetime
      
   two_log_rho_2 =  11.196 + LOG10(me*T*tau**2/my)
         
   two_log_rho = MINVAL([two_log_rho_1,two_log_rho_2])
      
   !      D_nl=zZ*zZ*6.*xn*xn*(xn*xn-xl*xl-xl-1.)
   !      cnl=9.933d-6*SQRT(mym)*(D_nl/SQRT(te))*ne*
   !     &     (11.538+LOG10(te/(D_nl*mym))+two_log_rho)
      
   D_nl = 6*n**2*(n**2 - l**2 - l - 1.)
      
   C_nl = cnl1*D_nl*(11.538 + LOG10(T/(D_nl*mym)) + two_log_rho)

   c2m(n-1) = C_nl  ! when l is at its maximum, C_nl = C_nl,nl-1
      
   c2p(n-2) = ((2.*l+1.)/(2.*(l-1.)+1.))*c2m(n-1)  ! detailed balance to get up rate C_nl,Cnl+1

   DO l = n-2, 1, -1    ! from n-2 and downwards
         
       tau = 1./anl(l)

       two_log_rho_2 =  11.196 + LOG10(me*T*tau**2/my)
         
       two_log_rho = MINVAL([two_log_rho_1,two_log_rho_2])
              
       D_nl = 6*n**2*(n**2 - l**2 - l - 1.)

       C_nl = cnl1*D_nl*(11.538 + LOG10(T/(D_nl*mym)) + two_log_rho)

       c2m(l) = C_nl - c2p(l)    ! C_nl = C_nl,l+1 + C_nl,l-1(=Cm(l))
                           
       c2p(l-1) = ((2.*l+1.)/(2.*(l-1)+1.))*c2m(l)  ! Detailed balance

   ENDDO

   l = 0
      
   cl(l,l+1) = (c1p(l)+c2p(l))*0.5  ! Average method 1 and 2 results....cl, c1p, c2p start at index 0
      
   DO l = 1, n-2
         
      cl(l,l-1) = (c1m(l)+c2m(l))*0.5 ! Average method 1 and 2 results
         
      cl(l,l+1) = (c1p(l)+c2p(l))*0.5 ! Average method 1 and 2 results
      
   ENDDO
      
   l = n-1
      
   cl(l,l-1) = (c1m(l)+c2m(l))*0.5

   pos1 = ((n-1)**2+(n-1))/2 + 1    !starting nl pos for given n

   Q(pos1:pos1+(n-1),pos1:pos1+(n-1)) = cl(0:n-1,0:n-1)   ! n number of l states for each n

   DEALLOCATE(anl)
   DEALLOCATE(c1p)
   DEALLOCATE(c1m)
   DEALLOCATE(c2p)
   DEALLOCATE(c2m)
   DEALLOCATE(cl)

END DO

END SUBROUTINE hI_coll_internal
