SUBROUTINE hI_freebound(atom, T, N, lambdavec, emissvec_integrated)
USE INTERFACE, ONLY : hI_recombspectrum
USE PARAMETERS, ONLY : h, c, eV, pi
USE M1, ONLY : atomm, lambdacut, myid
IMPLICIT NONE

TYPE(atomm), INTENT(IN) :: atom
DOUBLE PRECISION, INTENT(IN) :: T
INTEGER, INTENT(OUT) :: N
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE, INTENT(OUT) :: lambdavec, emissvec_integrated
! ------------------------------------------------
! TASK :    Compute the free-bound continuum emission for hydrogen. 
! INPUT :   atom (type atomm in M1)
!           T                     Temperature
!           N                     Number of wavelengths
! OUTPUT :  lambdavec             Wavelength vector
!           emissvec_integrated   Integrated free-bound emissivitties j_lambda*dlambda [erg s^-1 cm^3 ster^-1]
! GLOBALS:  lambdacut             Upper wl cutoff
! ------------------------------------------------

DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: emissvec_perA, emissvec_perE, CHI_EV, bftot
INTEGER :: i, tolevel, maxsize
DOUBLE PRECISION :: lambdamax, lambdamin, dL, dL_min, ans, E, nextE, dE

! -----------------------------------------------------------------------------------

lambdamin = 500.
lambdamax = maxval([1D6, lambdacut])     ! To 200 micron this routine is quick...alternative use lambdacut which is a global set in runparameters.dat (read in main.f90)
dL_min = 0.5

! ----------------------------------------------------------------------------------

N = INT(LOG(lambdamax/lambdamin)/LOG(1. + dL_min/lambdamin))  ! number of wavelength bins

ALLOCATE(lambdavec(N)); lambdavec = 0
ALLOCATE(emissvec_perA(N)); emissvec_perA = 0
ALLOCATE(emissvec_perE(N)); emissvec_perE = 0
ALLOCATE(CHI_EV(atom%N)); CHI_EV = 0
ALLOCATE(emissvec_integrated(N)); emissvec_integrated = 0
ALLOCATE(bftot(atom%N)); bftot = 0

lambdavec(1) = lambdamin

DO i = 2, N   ! Create log. spaced wavelength array

  lambdavec(i) = lambdavec(i-1)*(1. + dL_min/lambdamin)

END DO

CHI_EV = 13.595 - atom%LEVELS_E/eV                 ! Ionization energies en eV

DO tolevel = 1, atom%N

   DO i = 1, N-1     ! for each wavelength..

      E = h*c/(lambdavec(i)*1e-8 )/eV       ! photon energy in eV

      nextE = h*c/(lambdavec(i+1)*1e-8)/eV      ! Energy of next wavelength in array, Note this is a LOWER energy

      dE = E - nexte  ! width of energy bin [eV]...nexte is smaller
                
  ! DO tolevel = 1, atom%N  ! For each nl level in HI...can use whole atom (N_LEVELS solutions not needed)

      CALL hI_recombspectrum(tolevel, CHI_EV(tolevel), atom%LEVELS_G(tolevel), T, E, ans)   ! ans is [erg cm^3 s^-1 eV^-1]... NOT per steradian but total         

      bftot(tolevel) = bftot(tolevel) + ans*dE
      
     ! emissvec_perE(i) = emissvec_perE(i) + ans   ! add emisson to total energy per eV

     emissvec_perA(i) = 1./(4.*pi)*ans*h*c*1e8/(eV*(lambdavec(i))**2)      ! j_lambda = j_E dE/dlambda = j_E hc/lambda^2...All lengths in Å 1E-8/(1E-8)^2 = 1E8. This is per steradian now.
     emissvec_integrated(i) = emissvec_integrated(i) + emissvec_perA(i)*(lambdavec(i+1) - lambdavec(i))     ! Total flux in this wl bin..

  END DO
  

END DO

emissvec_integrated(N) = 0


END
