SUBROUTINE makealphavec(atom, T, N_levels, alpha_spec)
USE M1, ONLY : atomm, ngvec, Nelem, Nstages
USE INTERFACE, ONLY : hI_recomb_nl, feI_recomb, feII_recomb, oI_recomb, naI_recomb, mgI_recomb, makealpha 
IMPLICIT NONE
TYPE(atomm), INTENT(IN) :: atom
DOUBLE PRECISION, INTENT(IN) :: T
INTEGER, INTENT(IN) :: N_levels
DOUBLE PRECISION, DIMENSION(N_levels), INTENT(OUT) :: alpha_spec
! =================================================
! Task : Compute specific recombination rates at the specified temperature
! Input : atom        Atom to be computed (type atomm, M1)
!         T           Temperature
!         N_levels    Number of levels
! Output : alpha      Vector with recombination rates [cm^3 s^-1]
!          alpha_spec  The specific rates plus some allocation of rec to levels that are not included in atom
! Globals : ngvec   Number of levels in the ground multiplet (M1)
! =================================================
DOUBLE PRECISION, DIMENSION(29) :: HeIalphaIR, HeIbetaIR, HeIalphaIR5000
DOUBLE PRECISION, DIMENSION((Nstages-1)*Nelem) :: alphatot
INTEGER :: i, pos, allowed1, allowed2
DOUBLE PRECISION :: alphamissing, gtot, frac_groundstate, alphasum, t4
LOGICAL :: ATLEASTONEZERO
DOUBLE PRECISION, DIMENSION(N_levels) :: alphatrash
! --------------------------------------------------------------

alpha_spec = 0

frac_groundstate = 0.33   ! The fraction of the rec. rate allocated to the GS if specific rates are missing

! ---------------------------------------------------------------

ATLEASTONEZERO = .FALSE.       ! Important

WHERE (atom%RR_T0(1:N_levels) > 0)   ! If atom has recombination rates stored, use them

   alpha_spec(1:N_levels) = atom%RR_ALPHA0(1:N_levels)*(T/atom%RR_T0(1:N_levels))**(atom%RR_BETA(1:N_levels))

ELSEWHERE

   alpha_spec = 0

END WHERE

! Some elements have special routines to get the rec. rates:

IF (atom%NAME .EQ. ' HI  ') THEN

   CALL hI_recomb_nl(N_levels, T, alpha_spec)   ! Special routine for hydrogen

ELSE IF(atom%NAME .EQ. 'HeI  ') THEN

   continue   ! CARE

   !CALL readheI_ir(atom, HeIalphaIR, HeIbetaIR)

   !alpha_rad_notnormalized = alpha_rad_notnormalized + HeIalphaIR(1:N)*(T/1E4)**HeIbetaIR(1:N)

   !HeIalphaIR5000(1:N) =  HeIalphaIR(1:N)*(5000.0/10000.0)**(HeIbetaIR(1:N))

   !alpha_rad_notnormalized = alpha_rad_notnormalized + HeIalphaIR5000(1:N)*(T/5000.0)**(-0.75)

ELSE IF (atom%NAME .EQ. 'FeI  ') THEN

   CALL feI_recomb(T, atom%LEVELS_G(1:N_levels), N_levels, alpha_spec)   ! Special routine for FeI

ELSE IF ( atom%NAME .EQ. 'FeII ') THEN

   CALL feII_recomb(T, N_levels, alpha_spec)  ! Special routine for FeII

ELSE IF (atom%NAME .EQ. ' OI  ') THEN

   CALL oI_recomb(T,atom%LEVELS_G(1:N_levels), N_levels, alpha_spec)

ELSE IF (atom%NAME .EQ. 'NaI  ') THEN

   call naI_recomb(T, N_levels, alpha_spec)

ELSE IF (atom%NAME .EQ. 'MgI  ') THEN

   CALL mgI_recomb(T, N_levels, alpha_spec)

   t4 = T/1E4

   ! Dielectronic rec already included since I use Milne relations on TOPBASE cross sections.

   !alpha_spec(2) = alpha_spec(2) +  1./9.*1E-12*(0.51/t4 - 2.89 + 7.44*t4 - 0.72*t4**2)*t4**(-1.5)*exp(-2.41/t4)  ! Nussbaumer Storey 1986

   !alpha_spec(3) = alpha_spec(3) +  3./9.*1E-12*(0.51/t4 - 2.89 + 7.44*t4 - 0.72*t4**2)*t4**(-1.5)*exp(-2.41/t4)

   !alpha_spec(4) = alpha_spec(4) +  5./9.*1E-12*(0.51/t4 - 2.89 + 7.44*t4 - 0.72*t4**2)*t4**(-1.5)*exp(-2.41/t4)

END IF

IF (atom%CHARGE <= (Nstages-2) .and. atom%species_type .ne. 'mol') THEN     ! Ignore recombinations to highest ion stage            
   
	CALL makealpha(T, alphatot)   ! Obtain the TOTAL recombination rate (vector with all atoms) [cm^3 s^-1]
  
   pos = (Nstages-1)*(atom%DENSITY_ID-1) + 1 + atom%CHARGE  ! Find atom's position in the 2*Nelem vector alpha_rad
!    print*, atom%name
!    print*, pos
   alphamissing = alphatot(pos) - SUM(alpha_spec)    ! Total rec rate minus sum of known specific ones

   IF (alphamissing > 0) THEN        ! Recombination is missing..

      IF (alpha_spec(1) == 0.) THEN  ! If ground state has zero, allocate *frac_groundstate* of the missing rec. to it

         gtot = SUM(atom%LEVELS_G(1:ngvec(pos)))  ! ngvec

         alpha_spec(1:ngvec(pos)) = frac_groundstate*alphamissing*atom%LEVELS_G(1:ngvec(pos))/gtot  ! Share in prop. to stat weights

         alphamissing = alphamissing*(1.-frac_groundstate)  ! alphamissing now smaller..

      END IF

      IF (MINVAL(alpha_spec) == 0.) THEN  !If levels without rates exist, give them real rates to make up total..then all alfa_eff = alfa

         gtot = 0.

         DO i = 1, N_levels

            IF (alpha_spec(i) == 0.) THEN   ! allocate missing rec. only to levels which have zero rates so far

               gtot = gtot + atom%LEVELS_G(i)

            END IF

         END DO

         DO i = 1, N_levels

            IF (alpha_spec(i) == 0.) THEN   ! allocate in prop. to statistical weights

               alpha_spec(i) = alphamissing*atom%LEVELS_G(i)/gtot

            END IF

         END DO


      ELSE  !..otherwise add missing rate to top level

         !gtot = SUM(atom%LEVELS_G(1:N_levels))  !  marked out 4 lines

         !DO i = 1, N

         !   alpha_rad(i) = alpha(i) + alphamissing*atom%LEVELS_G(i)/gtot

         !END DO

         alpha_spec(N_levels) = alpha_spec(N_levels) + alphamissing  !  add all missing recombinations to highest state (makes sure that all recombination continuum energy is emitted)


      END IF


   ELSE IF (SUM(alpha_spec) > 10*alphatot(pos)) THEN ! The individual rates sum to more than the total rate by a too large factor : stop run

      WRITE(*,*) 'ERROR in makealphavec : Individual alphas > factor*alpha_tot. Atom:', atom%NAME, 'T:', T;PRINT *, 'Individual alphas:', sum(alpha_spec);PRINT *, 'Total alpha', alphatot(pos); stop 1

   ELSE   ! The individual rates sum to more than the total rate but not by too much : renormalize the specific rates so they match alpha_rad

      WRITE(*,*) 'WARNING in makealphavec : Individual alphas > alpha_tot (but < 2 alpha_rad). Renormalizing.', atom%NAME, T, sum(alpha_spec)/alphatot(pos)

      alphasum = sum(alpha_spec)

      alpha_spec = alpha_spec/alphasum*alphatot(pos)

   END IF

END IF

END SUBROUTINE makealphavec
