MODULE memalloc_shared

    IMPLICIT NONE

    ! Interface the memory sharing subroutines into one name (memshare)
    interface memshare
        module procedure memalloc_shared_1d
        module procedure memalloc_shared_2d
        module procedure memalloc_shared_3d
        module procedure memalloc_shared_4d
        module procedure memalloc_shared_real_2d
        module procedure memalloc_shared_real_3d
        module procedure memalloc_shared_double
        module procedure memalloc_shared_int
        module procedure memalloc_shared_int_1d
        module procedure memalloc_shared_logical
        module procedure memalloc_shared_logical_1d
    end interface

    private
    public :: memshare  ! This allows CALL memshare() in other files

CONTAINS

    ! 1-D array of doubles
    SUBROUTINE memalloc_shared_1d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(1)  ! dim of the array
        double precision, pointer, intent(INOUT) :: array(:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            windowsize = int(array_shape(1),MPI_ADDRESS_KIND)*8_MPI_ADDRESS_KIND ! 1d doesn't need products
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if
        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_1d

    ! 2-D array of doubles
    SUBROUTINE memalloc_shared_2d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(2)
        double precision, pointer, intent(INOUT) :: array(:,:) ! Dim of the array
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entries the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*8_MPI_ADDRESS_KIND ! Entries*8 for double
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if

        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_2d

    ! 3-D array of doubles
    SUBROUTINE memalloc_shared_3d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(3)  ! Dim of the array
        double precision, pointer, intent(INOUT) :: array(:,:,:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entries the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*8_MPI_ADDRESS_KIND ! Entries*8 for double
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if

        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_3d

    ! 4-D array of doubles
    SUBROUTINE memalloc_shared_4d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(4)  ! Dim of array
        double precision, pointer, intent(INOUT) :: array(:,:,:,:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entries the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*8_MPI_ADDRESS_KIND ! ! Entries*8 for DBLE
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if

        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_4d

    ! 2-D array of reals
    SUBROUTINE memalloc_shared_real_2d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(2)  ! Dimension of the array
        real, pointer, intent(INOUT) :: array(:,:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entries the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*8_MPI_ADDRESS_KIND ! Entries*8 for REAL
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if

        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_real_2d

    ! 3-D array of reals
    SUBROUTINE memalloc_shared_real_3d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(3)  ! Dimension of the array
        real, pointer, intent(INOUT) :: array(:,:,:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entires the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*8_MPI_ADDRESS_KIND ! Entries*8 for REAL
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if

        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_real_3d

    ! Single double
    SUBROUTINE memalloc_shared_double(comm, rank, value, win)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        double precision, pointer, intent(INOUT) :: value
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for the double, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            windowsize = 8_MPI_ADDRESS_KIND ! Single value => just DBLE
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if
        CALL C_F_POINTER(cf_pointer, value)
    END SUBROUTINE memalloc_shared_double

    ! Single integer
    SUBROUTINE memalloc_shared_int(comm, rank, value, win)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, pointer, intent(INOUT) :: value
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for an integer, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            windowsize = 4_MPI_ADDRESS_KIND ! Single value => just INT
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if
        CALL C_F_POINTER(cf_pointer, value)
    END SUBROUTINE memalloc_shared_int

    ! 1-D array of integers
    SUBROUTINE memalloc_shared_int_1d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(1)  ! Dim of the array
        integer, pointer, intent(INOUT) :: array(:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entires the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*4_MPI_ADDRESS_KIND ! Entries*4 for INT
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if
        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_int_1d

    ! Single logical
    SUBROUTINE memalloc_shared_logical(comm, rank, value, win)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        logical, pointer, intent(INOUT) :: value
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for the logical, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            windowsize = 4_MPI_ADDRESS_KIND ! Single LOGICAL
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if
        CALL C_F_POINTER(cf_pointer, value)
    END SUBROUTINE memalloc_shared_logical

    ! 1-D array of logicals
    SUBROUTINE memalloc_shared_logical_1d(comm, rank, array, win, array_shape)
        USE MPI
        USE ISO_C_BINDING

        IMPLICIT NONE

        ! Arguments
        integer, intent(IN) :: comm, rank
        integer, intent(IN) :: array_shape(1)  ! dim of the array
        logical, pointer, intent(INOUT) :: array(:)
        integer, intent(OUT) :: win
        ! Variables
        integer :: disp_unit = 1
        integer(KIND=MPI_ADDRESS_KIND) :: windowsize
        integer :: ier, byteval
        TYPE(C_PTR) :: cf_pointer

        !! Code to set up the shared memory for array, with array_shape, pointer cf_pointer
        !! with the right comm and rank, in a correct window
        if (rank == 0) then
            byteval = PRODUCT([array_shape])  ! Determine how many entires the array has
            windowsize = int(byteval,MPI_ADDRESS_KIND)*4_MPI_ADDRESS_KIND ! Entries*4 for LOGICAL
        else
            windowsize = 0_MPI_ADDRESS_KIND
        end if
        CALL MPI_BARRIER(comm, ier)

        disp_unit = 1
        CALL MPI_WIN_ALLOCATE_SHARED(windowsize, disp_unit, MPI_INFO_NULL, comm, cf_pointer, win, ier)

        if (rank /= 0) then
            CALL MPI_WIN_SHARED_QUERY(win, 0, windowsize, disp_unit, cf_pointer, ier)
        end if
        CALL C_F_POINTER(cf_pointer, array, array_shape)
    END SUBROUTINE memalloc_shared_logical_1d
END MODULE
