SUBROUTINE regemorter(N, T_typical, atom)
USE M1, ONLY : atomm
USE PARAMETERS, ONLY : k, c, h, pi
USE FILES, ONLY : file_regemorter
IMPLICIT NONE
INTEGER, INTENT(IN) :: N
DOUBLE PRECISION, INTENT(IN) :: T_typical
TYPE(atomm), INTENT(INOUT) :: atom
! ==================================
! Task  :  Compute collision strengths for allowed lines using Regemorter's formula (Regemorter1962). Also computes forbidden lines. NOTE currently another treatment is implemented, C = 0.004 g1*g2
! Input :  N             Number of levels 
!          T_typical     Estimate of temperature [K] (Effective collisions strengths have a weak temp. dependence, so exact values does not have to be used)
!          atom          An atomm object (module M1) with line data stored in atom%LAMBDA, atom%A and atom%LEVELS_G
! Output : The atom%UPSILON0 array of collision strengths is filled.
! ==================================
INTEGER :: i, j, bin, rot_range, v_band,v_band2
DOUBLE PRECISION :: f_osc, gf, dE, y, Py, fuse, Ause, lambda, fosc
DOUBLE PRECISION, DIMENSION(10) :: yvec, Pyvec_n, Pyvec_i
DOUBLE PRECISION, EXTERNAL :: E1
! -----------------------------------------------------------

yvec = [0.01, 0.02, 0.04, 0.1, 0.2, 0.4, 1., 2., 4., 10.]   ! Regemorter1962 Table 2, column 1

Pyvec_n = [1.16, 0.956, 0.758, 0.493, 0.331, 0.209, 0.100, 0.063, 0.040, 0.023]  ! Regemorter1962 Table 2, column 2

Pyvec_i = [1.16, 0.977, 0.788, 0.554, 0.403, 0.290, 0.214, 0.201, 0.200, 0.200]  ! table 2, column 3

!OPEN(10, file=TRIM(file_regemorter)//atom%NAME, STATUS='REPLACE') ! These are not needed

!!------------------------------------------------------
!! sl: special case for molecules used in Liljegren+20, removed for real data instead
!IF (atom%species_type .EQ. 'mol') THEN
!! CARE special case for CO, where all the collision strenghts
!! of CO are set to be 0.004*g1*g2
!    

!        DO j = 2, N  ! For each upper level..
!           DO i = 1, (j-1)  ! for each lower level..
!                IF (atom%UPSILON0(j,i) .EQ. 0.0) THEN
!!                    atom%UPSILON0(j,i) = 0.004*atom%LEVELS_G(i)*atom%LEVELS_G(j)
!                    atom%UPSILON0(j,i) = 0.0
!                ENDIF
!            END DO
!        END DO


!------------------------------------------------------
!------------------------------------------------------

IF (atom%species_type .EQ. 'mol') THEN

! CARE special case for CO, where all the collision strenghts
! of CO are set to be 0.004*g1*g2
!    atom%UPSILON0(j,i) = 0.0
if (atom%UPSILON0_EXISTS .eqv. .FALSE.) then
        DO j = 2, N  ! For each upper level..
           DO i = 1, (j-1)  ! for each lower level..
                atom%UPSILON0(j,i) = 0.004*atom%LEVELS_G(i)*atom%LEVELS_G(j)
            END DO
        END DO
end if

!! FIXME trying to fix the convergence issues
!else
!DO j = 2, N  ! For each upper level..
!           DO i = 1, (j-1)  ! for each lower level..
!                atom%UPSILON0(j,i) = 0.004*atom%LEVELS_G(i)*atom%LEVELS_G(j)
!            END DO
!        END DO
!end if

ELSE
DO j = 2, N  ! For each upper level..

   DO i = 1, (j-1)  ! for each lower level..

      IF (atom%UPSILON0(j,i) .EQ. 0) THEN  ! If no collision strength exists..
      
         !f_osc = atom%A(j,i)/6.67e13*atom%LEVELS_G(j)/atom%LEVELS_G(i)*(atom%LAMBDA(j,i)/10.)**2

         !dE = atom%LEVELS_E(j) - atom%LEVELS_E(i)  ! Shuld be in ergs

         !CALL gaunt_regemorter(dE, i, j, T_typical, gf)

         !atom%UPSILON0(j,i) = 197.33*f_osc*gf*atom%LEVELS_G(i)/(atom%LEVELS_E(j)/eV - atom%LEVELS_E(i)/eV)   ! Source?

         y = (atom%LEVELS_E(j)-atom%LEVELS_E(i) + 1E-18)/(k*T_typical)  ! y = DeltaE/kt. Add small number (1E-18) to avoid zero when levels_e(i) = levels_e(j)

         lambda = h*c/(atom%LEVELS_E(j) - atom%LEVELS_E(i) + 1E-18)   ! cannot use atom%LAMBDA, transitions with zero A values have atom%lambda = 0

         CALL findbin2(10, yvec, y, bin)  ! find which bin in the yvec array y is

         IF (atom%CHARGE .EQ. 0) THEN  ! neutral atom..

            IF (bin .ne. 0) THEN

               Py = Pyvec_n(bin)

            ELSE IF (y > yvec(10)) THEN

               Py = 0.066*y**(-0.5)        ! Regemorter1962 table 2, col 2   also Axelrod1980  4.50

            ELSE IF (y < yvec(1)) THEN

               Py = SQRT(3.)/(2*pi)*E1(y)  ! Regemorter1962  table 2, col 2

            END IF

         ELSE IF (atom%CHARGE >= 1) THEN   ! ion..

            IF (bin .ne. 0) THEN

               Py = Pyvec_i(bin)

            ELSEIF (y > yvec(10)) THEN

               Py = 0.2           ! Regemorter1962 table 2 col 3   also Axelrod1980 4.51

            ELSEIF (y < yvec(1)) THEN

               Py = SQRT(3.)/(2*pi)*E1(y)  ! Regemorter1962    table 2, col 3

            END IF

         END IF

         fosc = 1.49*atom%LEVELS_G(j)/atom%LEVELS_G(i)*lambda**2*atom%A(j,i)    ! Rutten eq 2.68..lambda in cm here. 1/6.67e13*1e14 (nm)= 1.49

         IF (Py > 0 .AND. fosc > 1e-3) THEN   ! Allowed transition, limit from rutten sec 3.2 page 52

            atom%UPSILON0(j,i) = 2.387E6*lambda**3*atom%A(j,i)*atom%LEVELS_G(j)*Py  ! Regemorter1962 eq 18  A-value is prop. to levels_g(i)

         ELSE IF (Py > 0 .AND. fosc <= 1e-3) THEN  ! Forbidden transition, use f_osc = 1E-2

            fuse = 1E-2*atom%LEVELS_G(j)  ! CARE  absorption osc. strength scales with stat. weight of upper level (j)

            Ause = 0.667*atom%LEVELS_G(i)/atom%LEVELS_G(j)*fuse/lambda**2     ! Rutten eq 2.68  lambda is in cm, when in nm coefficent is 6.67E13 --> divide by 1E14 --> 0.667

            atom%UPSILON0(j,i) = REAL(MAXVAL([0.1d0*atom%LEVELS_G(i)*atom%LEVELS_G(j), 2.387E6*lambda**3*Ause*atom%LEVELS_G(j)*Py]))   ! if value below 0.1 use 0.1
    
            IF (atom%UPSILON0(j,i) > 10.*atom%LEVELS_G(i)*atom%LEVELS_G(j)) atom%UPSILON0(j,i) = 10.*atom%LEVELS_G(i)*atom%LEVELS_G(j)   ! CARE avoid crazily large values for very small energy separations
            ! special case for CO to be comparable to the cross section from data - SL
!            IF (atom%NAME .EQ. 'CMI  ') THEN
!            atom%UPSILON0(j,i) = 0.002*atom%LEVELS_G(j)
!            ELSE 
            IF (atom%Z >= 1) THEN

               atom%UPSILON0(j,i) = 0.004*atom%LEVELS_G(i)*atom%LEVELS_G(j)  ! CARE use 0.004 from Axelrod 

            END IF

         END IF
        
      END IF

      !WRITE(10,'(2I5,3ES10.3)') j, i, atom%UPSILON0(j,i)

   END DO

END DO
END IF
!close(10)

END


!SUBROUTINE GAUNT_REGEMORTER(dE, i, j, T, gf)
!USE M1
!USE PARAMETERS
!IMPLICIT NONE

!DOUBLE PRECISION, INTENT(IN) :: dE
!INTEGER, INTENT(IN) :: i, j
!DOUBLE PRECISION, INTENT(IN) :: T
!DOUBLE PRECISION, INTENT(OUT) :: gf

!DOUBLE PRECISION :: y, expintval

! ----------------------------------------------

!y = dE/(k*T)

!call e1xa(y, expintval)

!gf = 3*sqrt(3.)/(2*pi)*(1 - y*exp(y)*expintval)                 ! Woolley-Allen         (see Regemorters article)
                                                                ! Temp                                                                ! Regemorter has  better fit

!END

FUNCTION E1(X)
!
!C       ============================================
!C       Purpose: Compute exponential integral E1(x)
!C       Input :  x  --- Argument of E1(x) 
!C       Output:  E1 --- E1(x) ( x > 0 )
!C       ============================================
!C
        IMPLICIT DOUBLE PRECISION (A-H,O-Z)

        IF (X.EQ.0.0) THEN
           E1=1.0D+300
        ELSE IF (X.LE.1.0) THEN
           E1=-DLOG(X)+((((1.07857D-3*X-9.76004D-3)*X+5.519968D-2)*X - 0.24991055D0)*X+0.99999193D0)*X-0.57721566D0
        ELSE
           ES1=(((X+8.5733287401D0)*X+18.059016973D0)*X + 8.6347608925D0)*X+0.2677737343D0
           ES2=(((X+9.5733223454D0)*X+25.6329561486D0)*X + 21.0996530827D0)*X+3.9584969228D0
           E1=DEXP(-X)/X*ES1/ES2
        ENDIF
END
