import glob
import subprocess
from shutil import copyfile
import math
import numpy as np
import time


#Base name for model, to be used when saving across several days
model_ID = 'onezonesolar_005M_005v_highalpha_steady'

#Runfile name
runfile = 'playmodel_knatest'

#Read in number of zones and explosion file name from runfile
expfile = np.loadtxt(runfile,dtype=str,skiprows=3,max_rows=1,unpack=True)[1]
expfile = expfile.replace("'","")
zone_nr = int(np.loadtxt(runfile,dtype=str,skiprows=4,max_rows=1,unpack=True)[1])

#Write model file containing run information
runinfo = open("runinfo.txt","w")
runinfo.write('Run name: %s \n'%model_ID)
runinfo.write('Runfile: %s \n'%runfile)
runinfo.write('Number of zones: %d \n'%zone_nr)
runinfo.write('Explosion Model: %s \n'%expfile)
runinfo.write('Mej and Vej: 0.05Msol 0.05c \n')
runinfo.write('Thermalization: Kasen, Q_A = 0.1 Q_B \n')
runinfo.write('Recomb. Alpha value: 1e-11 \n')
runinfo.write('Notes:\n')
runinfo.close()


#Grid info

#Elements: Ce, Te
#Mass grid: 0.01, 0.05, 0.1
#Velocity grid: 0.03, 0.1, 0.2
#Epoch: 100.0, 30.0, 10.0, 3.0
#Recomb: 1e-11, 1e-13

#Read in Mej and v_ej from explosion file
data = np.loadtxt(expfile,dtype=str,skiprows=1,max_rows=5,unpack=True)

Mzone, vzone = np.array([]),np.array([])
for i in range(1,zone_nr+1): #skip first column of row
    Mzone = np.append(Mzone,float(data[i,0])) #zone mass in Msol
    vzone = np.append(vzone,float(data[i,2])/3.0E5) #velocity of exterior zone border in units of c

Mej = np.sum(Mzone) #total ejecta mass in units of Msol
v_ej = vzone[zone_nr - 1] #ejecta velocity at edge of exterior zone, units of c

Mzone_perc = Mzone/Mej #zone masses in percent of total mass

#Energy deposition
def dep_energy(t,Mej,v_ej):
    '''
       Input: t must be in days, total ejecta mass in Msol, maximum ejecta velocity in units of c
       Return: total radioactive energy deposition in erg/s
    '''
    alpha = 1.3 #power law slope for beta decay radioactive power distribution
    energydep_1d = 1.0E10 #energy dep. at 1 day in erg/s/g
    
    #Barnes 2016 formalism in erg/s/g * Mej * f(t)
    Barnes =  (1.0E10*(t**-alpha)*(Mej*2.0E33)) * (3.6E-1*(math.exp(-0.27*t) + ( math.log1p(0.2*t**(0.6)) )/( 0.2*t**(0.6) ) ) )

    #Kasen 2019 formalism in erg/s/g * Mej * f(t), including alpha and beta decay, total should add to 1e10 erg/s/g
    
    t_e = 12.9 *((Mej/0.01)**0.66) * ((v_ej/0.2)**-2.0) #Compton electron thermalization timescale in days
    t_y = 0.3 * ((Mej/0.01)**0.5) * ((v_ej/0.2)**-1.0) #Gamma ray thermalisation timescale in days
    
    Q_B = 0.9 * energydep_1d * (t**-alpha) # Beta decay radioactive power
    Q_A = 0.1 * Q_B  # Alpha decay radioactive power (Q_A = 0.05 - 0.40 Q_B)

    f_B = 0.2*(1 + (t/t_e))**(-1.5) + 0.5*(1.0 - math.exp(-1.0*(t_y**2.0)/(t**2.0))) #fraction of absorbed beta decay energy
    f_A = (1 + (t/(3*t_e)))**(-1.5) #fraction of absorbed alpha decay energy


    Kasen = (Mej*2.0E33) * ((f_B * Q_B) + (f_A * Q_A))
     
    return Kasen


# Days, in desired run order
#days = [5.8,6.4,7.1,7.8,8.6,9.4,10.4,11.4,12.5,13.8,15.0,16.5,18.2,20.0,22.0,24.2,26.6,30.0,33.0,36.3,40.0,44.0,48.0,52.0,57.2,63.0,69.3,75.9,83.6,92.0,100.0]

#days = [10.4,11.4,12.5,13.8,15.0,16.5,18.2,20.0,22.0,24.2,26.6,30.0,33.0,36.3,40.0,44.0,48.0,52.0,57.2,63.0,69.3,75.9,83.6,92.0,100.0]

#days = [92.0,100.0]

days =[6.4,5.8,5.3]

#Dynamic day and energy dep. placement in ip and deposition files
for d in days:
    energy = dep_energy(d,Mej,v_ej)
    print('------------------')
    print('model running at %f days'%d)
    print('Energy deposition is %5.2e'%energy)
    print('------------------')
    
    lines = open(runfile).read().splitlines()
    lines[0] = lines[0].replace('xxx', str(d))
    open('playmodel_temp','w').write('\n'.join(lines))

    depfile = open("out/main/dep.dat","w")
    depfile.write('  %5.2e \n'%energy)
    for i in range(1,zone_nr+1):
        zone_energy = energy * Mzone_perc[i-1]
        depfile.write('   %i \t %5.2e \t %5.2e \n'%(i,zone_energy,Mzone_perc[i-1]))
    depfile.close()

    
#    if (d == days[0]):
        #Initialise process only for the first day CARE do not use if using previous solution from past run as starting point
#        subprocess.call('bash iterate.sh 1 2 64 highsteady_ playmodel_temp n n 500 50000 1 1000000 100 100 100', shell=True)

    #Use previous solution as starting guess
    subprocess.call('bash iterate.sh 200 2 64 highsteady_' + str(d) + ' playmodel_temp n y 500 50000 1 1000000 100 100 100', shell=True)

    #Save all relevant output files before moving to next day
    subprocess.call('bash putfiles.sh ' + model_ID + '_' + str(d) + ' ' + str(zone_nr) + ' ' + runfile, shell=True) #TEMP copy timedays and teqsordered to out directory for time dependent process
    
    time.sleep(1.0)


