SUBROUTINE statisticalandthermalequilibrium(N_atoms, atoms, R_index, DIMS, x, abinvec, fground, nmolvec, EXCmax, dep_tot_pervol, radheating, IONRATES, alfa_dust, tau_tresh, abintresh, T_prev, xe_prev, T_eq, EXC_MATRIX, cool_co, cool_sio, fv1_co, fv1_sio, ntheatrate,convprobs)
use M1, ONLY : Nmol, Npis, Nabin, ITERATIVERUN, R_index_str, ngvec, atomm, myid, globid, IDP, map2, time, time_prev, lambdaion, SFMAX, Nelem, Nmol, DOPRINT, MOLECULESWITCH, TIMEDEPENDENT_T, Nstages
use INTERFACE, ONLY : alfa_photoionization, stateq, netheating, createtau_s
use PARAMETERS, ONLY : eV, h, c, k
USE FILES
USE MPI
IMPLICIT NONE
INTEGER :: N_atoms, DIMS, EXCmax
INTEGER, INTENT(IN) :: R_index
TYPE(atomm), DIMENSION(N_atoms) :: atoms
DOUBLE PRECISION, DIMENSION(DIMS) :: x
DOUBLE PRECISION, DIMENSION(Nabin), INTENT(INOUT) :: abinvec, fground
DOUBLE PRECISION, DIMENSION(maxval([1,Nmol])), INTENT(IN) :: nmolvec
REAL, DIMENSION(:,:) :: IONRATES  ! Nabin*Npis
DOUBLE PRECISION :: dep_tot_pervol
DOUBLE PRECISION, INTENT(IN) :: radheating, alfa_dust, tau_tresh, abintresh, T_prev, xe_prev
DOUBLE PRECISION, INTENT(OUT) :: T_eq, cool_co, cool_sio, fv1_co, fv1_sio
DOUBLE PRECISION, DIMENSION(N_atoms, EXCmax), INTENT(OUT) :: EXC_MATRIX
DOUBLE PRECISION, INTENT(OUT) :: ntheatrate
LOGICAL, DIMENSION(N_atoms+1), INTENT(OUT) :: convprobs

! ===========================================================
! Task : Solves for statistical (excitation structure) and thermal equilibrium for a given zone. Ionization balance assumed given.
!
! Input : N_atoms        Number of atoms
!         atoms          Array of atoms (type atomm (module M1))
!         R_index        Shell number. Needed to file ionization rates for later use.
!         DIMS           Dimension of x array (contains state variables, see modelreader.f90)
!         x              Array with basic zone data
!         abinvec        Fractional abundaces in shell
!         EXCmax         Maximum atom size to solve for
!         dep_tot_pervol Energy deposited in the shell per unit volume and time  [erg cm^-3 s^-1]
!         radheating     Radiative heating (from previous iteration) åerg cm^-3 s^-1]
!         IONRATES       Ionization rates, from excited states needed only [s^-1]
!         alfa_dust      Dust opacity [cm^-1]
!         tau_tresh      Limiting line opacity. Used to find maximum lower level with relevant optical depth.
!         abintresh      Limiting abundance to include in Spencer-Fano solver.
!         T_prev         Previous temperature solution for zone, used in time dep. mode
!         xe_prev        Previous electron fraction for zone, used in time dep. mode
!
! Output: T_eq           Equilibrium temperature [K]
!         EXC_MAT        Equilibrium excitation structures [cm^-3]
!         cool_co        Cooling by CO [erg s^{-1} cm^{-3}]
!         cool_sio       Cooling by SiO [erg s^{-1} cm^{-3}]
!         fv1_co         Fraction of CO cooling done by fundamental band
!         fv1_sio        Fraction of SiO cooling by fundamental band
!
! GLOBALS:  Nmol        Number of molecules (set in M1)
!           Nabin       Number of abin elements (set in M1)
!           Npis        Number of states which photoionizations occur from (set in M1)
!
! Implicit output : atoms%N_LEVELS_ABSORB      Maximum level that has lines with tau > tautresh.
! ============================================================
DOUBLE PRECISION, DIMENSION(Nabin,EXCmax) :: CTINMAT, DIRECTIONMAT, TOTINMAT
DOUBLE PRECISION, DIMENSION(4) :: pio
TYPE(atomm) :: atom
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: levels_pop
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: TAU_S
INTEGER :: m, niterT, i, j, N_levels, l, u, lmax, umax, abin, level, nitermax, nn, jj
DOUBLE PRECISION :: n_TOT, T, ne, xe, volume, heatrate, convlimT, gprim, dT, t2factor, Tuse
LOGICAL :: CONVERGED, FOUND
!INTEGER, PARAMETER :: N_CHANNELS = 262      ! Maximum number of excitation channels in sf
DOUBLE PRECISION :: frac_heat, np, nh1, nh2, T2, T_old, gsum, lambdac_max, lambdac_min, f, z, trash1, trash2, trash3, trash4, sftot, adcool, damp, jval, cttemp, g1, g2, apa
DOUBLE PRECISION, DIMENSION(Nabin) :: fracvec_exc, fracvec_ion, vec_Xeff, abinvec_use
DOUBLE PRECISION, DIMENSION(Nabin, SFMAX) :: fracmat_exc
INTEGER, DIMENSION(Nabin) :: nlx
INTEGER :: nik0, failsig, FAIL, counter, info, ierr, FH
DOUBLE PRECISION, DIMENSION(:,:), ALLOCATABLE :: ALFA_MAT
DOUBLE PRECISION, DIMENSION(Nabin,maxval([Npis,ngvec])) :: EXC_MAT_ABIN
INTEGER, PARAMETER :: Nlambda=1000
DOUBLE PRECISION, DIMENSION(Nlambda) :: lambdavec, alfa_c
DOUBLE PRECISION, DIMENSION(Nlambda, (Nstages-1)*Nelem+Nmol+3) :: alfa_c_ind
DOUBLE PRECISION, DIMENSION((Nstages-1)*Nelem) :: gammant, nabsnt      ! Nonthermal ionization rates, later needed for ion. sol.
DOUBLE PRECISION, DIMENSION(N_atoms, EXCmax) :: EXC_MATRIX2
DOUBLE PRECISION, DIMENSION(Nabin) :: ndensvec_abin=0    ! used only for bremsstrahlung opacity, assume zero here
REAL, DIMENSION(Nabin,Npis) :: IONRATES_COLLION, IONRATES_USE, IONRATES_COLLION2
! -------------------------------------------------------------------------------------------------------------------

nitermax = 30  ! Maximum number of iterations for thermal equilibrium. If exceeding this, last solution is used.

damp = 0.8 ! CARE     ! 1=normal step. Damping for Newton-Raphson step in thermal eq.

convlimT = 1E-3   ! CARE Convergence criteria for temperature Tnew/Told < convlimT

t2factor = 1.01  !TEMP  ! Use 1% step size for thermal derivatives.

lambdac_max = 5000.   ! Maximum wavelength for continuum absorption (used to comp. cont. destr. prob. only)

lambdac_min = 100.    ! Smallest wl for continuum absorption

!-----------------------------------
CTINMAT = 0  ! Important
DIRECTIONMAT = 0
TOTINMAT = 0
IONRATES_COLLION = 0.
IONRATES_COLLION2 = 0.

nn = maxval([Npis,ngvec])
ALLOCATE(ALFA_MAT((Nstages-1)*Nelem+Nmol+3,nn)); ALFA_MAT=0

alfa_c = 0

nik0 = 1500                   ! Number of energy bins used by sf for the non-thermal electron spectrum.

EXC_MATRIX = 0                        ! Important

EXC_MATRIX2 = 0  ! Imp

fracvec_exc = 0

fracmat_exc = 0

frac_heat = 0

CONVERGED = .FALSE.  ! Imp

failsig = 0

niterT = 0

EXC_MAT_ABIN = 0

! -------------------------------------------------------------------------------------------------------------------

volume          = x(DIMS-8)             ! Volume of each shell at current time [cm3]

n_TOT           = x(DIMS-3)             ! Total number density [cm^-3]

xe              = x(DIMS-2)		! Electron fraction


T               = x(DIMS)		! Temperature (at current time) [K]

T_old = T

Tuse = T   ! Used for sf

ne = n_TOT*xe  ! Electron density [cm^-3]

nik0 = 2000   ! CARE

nlx = 0       ! CARE 

! Put all atoms in their ground states for initial computations of the continuum destruction probabilities:

DO i = 1, N_atoms

    EXC_MAT_ABIN(atoms(i)%ABIN,1) = x(atoms(i)%POS)  ! Total number density of this atom or molecule

END DO

f = (lambdac_max/lambdac_min)**(1./DBLE(Nlambda))  ! Relation between neighboring wavelengths f = lambda(i+1)/lambda(i)

lambdavec(1) = lambdac_min

DO i = 2, Nlambda   ! Create wavelength vector

   lambdavec(i) = lambdavec(i-1)*f

END DO

IF (globid == IDP) PRINT *, 'Entering sf..'
!print*, 'abinmat CO!', abinvec(106), fground(106)
abinvec_use = abinvec*fground  ! assume only ionizatons from ground state can occur




WHERE (abinvec_use < abintresh)   ! Ignore computations for elements with low abundances to speed up sf. abintresh set by sf_tresh in ip file.  NT ionization rates for these are later approximated in ionization.f90

   abinvec_use = 0.

END WHERE

! Mark in some elements that are useful as 'templates' for eg molecules and trace elements. However,
! fraction cannot be too small, then sf crashes. 1e-6 seems ok

if (abinvec_use(4) > 1e-5) abinvec_use(4) = abinvec(4)  ! ..but always include CI beause that rate is used for molecules (CO and SiO)

if (abinvec_use(54) > 1e-5) abinvec_use(54) = abinvec(54)   ! and always include Fe I because that is used as template for trace elements in ionization.f90

if (abinvec_use(55) > 1e-5) abinvec_use(55) = abinvec(55)   ! and always include Fe II because that is used as template for trace elements in ionization.f90

if (abinvec_use(19) > 1e-5) abinvec_use(19) = abinvec(19)  ! Ne II

if (abinvec(126) > 1E-8) abinvec_use(126) = abinvec(126) ! TEMP
if (abinvec(131) > 1E-8) abinvec_use(131) = abinvec(131) ! TEMP
if (abinvec(136) > 1E-8) abinvec_use(136) = abinvec(136) ! TEMP
if (abinvec(141) > 1E-8) abinvec_use(141) = abinvec(141) ! TEMP

abinvec_use = abinvec_use/sum(abinvec_use)  ! Normalize (dont think its necessary though)

! Call Spencer-Fano routine to see what fractions go into heating, ionizations and excitations

if (globid == 2) then

  print *, 'input to sf', globid
  do i = 1, Nabin
   print *, i, abinvec_use(i)
  end do

   print *, 'xe', xe

end if

abinvec_use(70) = 0 !Don't include the protons

CALL sf(xe, abinvec_use, 1, 1, failsig, frac_heat, fracvec_exc, fracmat_exc, fracvec_ion, pio, vec_Xeff, nlx, nik0, R_index,REAL(Tuse),Nstages) ! Third argument:include excitations?

if (doprint .and. (globid == IDP)) then

   print *, 'sf finished'

  continue

end if

where (isnan(fracmat_exc))  ! CARE

   fracmat_exc = 0

end where

where (isnan(fracvec_exc))  ! CARE

   fracvec_exc = 0

end where

! Make sure all channels add up to 1

sftot = frac_heat + sum(fracvec_exc) + sum(fracvec_ion) - fracvec_ion(14) + sum(pio)

frac_heat = frac_heat/sftot  ! Normalize

fracmat_exc = fracmat_exc/sftot

fracvec_ion = fracvec_ion/sftot

pio = pio/sftot

OPEN(10,file=TRIM(file_sf)//R_index_str(R_index),STATUS='REPLACE')  ! write SF output

  WRITE(10,'(A20, 3ES10.2)') 'Heat  Ion Exc', frac_heat, SUM(fracvec_ion), SUM(fracvec_exc)

  DO i = 1, Nabin

     write(10,*) 'i  ion  exc'

     WRITE(10,'(I10,2ES10.2)') i, fracvec_ion(i), fracvec_exc(i)

     DO j = 1, SFMAX

        IF (fracmat_exc(i,j) > 0) THEN

           WRITE(10,'(I10,ES10.2)') j, fracmat_exc(i,j)

        END IF

     END DO

  END DO

close(10)

ntheatrate = dep_tot_pervol*frac_heat       ! Energy going to heating [ergs/s/cm3]

heatrate = ntheatrate + radheating   ! Total heating rate (nonthermal + radiative) per cm3

OPEN(10,file=TRIM(file_heatfracs)//R_index_str(R_index),STATUS='REPLACE')  ! Not further used

  WRITE(10,'(I5,ES12.2,A10,ES12.3,A10,ES12.3)') R_index, heatrate, 'NT:', ntheatrate/heatrate, 'RAD:', radheating/heatrate

CLOSE(10)

! Store nonthermal ionization rates in files for later use in ionization computation:

OPEN(30,file=TRIM(file_gammant)//R_index_str(R_index),STATUS='REPLACE')  ! Used by ionization.f90

OPEN(40,file=TRIM(file_nabsnt)//R_index_str(R_index),STATUS='REPLACE')  ! Used by radmaker.f90

! Compute non-thermal ionization rates for the ground states in all atoms

DO i = 1, (Nstages-1)*Nelem

   IF ((i .NE. 2) .and. (map2(i) .NE. 0)) THEN   ! Not for protons..

      IF (abinvec(map2(i)) > 0) THEN  ! If element i exists in this zone...

         IF (map2(i) .eq. 14) THEN  ! TEMPfix hardcoding OI special treatment because ionization to excited states

            gammant(i) = dep_tot_pervol*pio(1)/(n_TOT*abinvec(map2(i))*13.618*eV) + dep_tot_pervol*pio(2)/(n_TOT*abinvec(map2(i))*(13.618+3.325)*eV) + dep_tot_pervol*pio(3)/(n_TOT*abinvec(map2(i))*(13.618+5.018)*eV) + dep_tot_pervol*pio(4)/(n_TOT*abinvec(map2(i))*(13.618+14.870)*eV)     ! pio contains fraction of energy going into excited state ionizations...dividide by E_ion + E_exc to get rate

            ! inflow rates to the excited states in OII

            DIRECTIONMAT(15,2) = dep_tot_pervol*pio(2)/((13.618+3.325)*eV)*6./10.  ! OII note these are per unit volume and so dont divide with n_TOT

            DIRECTIONMAT(15,3) = dep_tot_pervol*pio(2)/((13.618+3.325)*eV)*4./10.

            DIRECTIONMAT(15,4) = dep_tot_pervol*pio(3)/((13.618+5.018)*eV)*4./6.

            DIRECTIONMAT(15,5) = dep_tot_pervol*pio(3)/((13.618+5.018)*eV)*2./6.

            DIRECTIONMAT(15,6) = dep_tot_pervol*pio(4)/((13.618+14.870)*eV)*6./12.

            DIRECTIONMAT(15,7) = dep_tot_pervol*pio(4)/((13.618+14.870)*eV)*4./12.

            DIRECTIONMAT(15,8) = dep_tot_pervol*pio(4)/((13.618+14.870)*eV)*2/12.


         ELSE

            gammant(i) = dep_tot_pervol*fracvec_ion(map2(i))/(n_TOT*abinvec(map2(i))*h*c/(lambdaion(i)*1E-8)) ! Ionization rate per particle and second...take total energy that went to ionizing element i per unit volume, divide by the number density of the element to get energy per particle and second, and divide by the ionization potential of the element to get the rate : ionizations per particle per second. Note we dont want to use the fground factor here : its not the ionization rate PER STATE but PER ATOM/ION

         END IF

         WRITE(30,'(2I5,ES10.2)') R_index, i, gammant(i)

         nabsnt(i) = gammant(i)*n_TOT*abinvec(map2(i))*volume      ! Total number of ionizations per sec in entire zone

         ! Find where in atoms vector element i resides.

         FOUND = .FALSE.

         counter = 0

         DO WHILE (FOUND .EQV. .FALSE.)

            IF (atoms(counter+1)%ABIN .EQ. map2(i)) THEN  ! map2 contains the ABIN codes sorted in 1-30 order (HI,HII,HeI,HeII,CI,CII..)

               FOUND = .TRUE.

            END IF

            counter = counter + 1

            IF (counter >= N_atoms) EXIT

         END DO

         IF (FOUND) THEN

            !gsum = SUM(atoms(counter)%LEVELS_G(1:ngvec(i))) ! Add up the statistical weights of the ground multiplet, vector ngvec contains the number of states in the ground term for all elements

            !DO j = 1, ngvec(i)      ! Ionization from all levels in ground multiplet : allocate number of ionizations w stat. weights

             WRITE(40,'(2I5,ES10.2)') R_index, i, nabsnt(i)

            !END DO

         END IF

      END IF

   END IF

END DO

CLOSE(30)

CLOSE(40)

IF (ITERATIVERUN) THEN    ! Add charge tranfer to excited states

   OPEN(10,file=TRIM(file_cttoexc)//R_index_str(R_index),STATUS='OLD')

   DO

      READ(10,*,END=123) abin, level, cttemp   ! cttemp is zero if CTON is set to 0 in runparameters.dat

      CTINMAT(abin,level) = CTINMAT(abin,level) + cttemp  ! The same level can be filled by several reactions

   END DO

123 close(10)

END IF

TOTINMAT = CTINMAT + DIRECTIONMAT  ! total direct level feeding: charge transfer plus direct ionizations (recombinations are added later)

open(10,file=TRIM(file_molcool2)//R_index_str(R_index), STATUS='REPLACE')  ! Default no mol cooling...zones with molcooling replace later

   write(10,*) R_index, 0

close(10)

DO i = 1, Nlambda  ! Create continuum opacity vector alfa_c (ALFA_MAT not used)..this should strictly be done anew in each iteration (new EXC_MAT_ABIN values), but since alfa_c has a very small influence we do only once

      CALL alfa_photoionization(lambdavec(i), N_atoms, atoms, REAL(EXC_MAT_ABIN), nmolvec, alfa_dust, T, ne, ndensvec_abin, alfa_c(i), ALFA_MAT)

      alfa_c_ind(i,:) = sum(ALFA_MAT,2)

END DO

DO WHILE (CONVERGED .EQV. .FALSE.)   ! Iterate for thermal equilibrium..

   IF (heatrate < 0.) THEN  ! something weird, put to low temperature and continue

     T = 0.9*T_old

     CONVERGED = .TRUE.  ! CARE   dont put outside, loop has to go at least once

   END IF

   IF (MOLECULESWITCH == 1 .AND. R_index .EQ. 3) THEN   ! CARE MOLECULESON means molecules are explicitly computed

      !   T = 500

      T = 2000. - 4*(time/(24.*3600)-259.)

      IF (T < 500.) T = 500.

      CONVERGED = .TRUE.

   ELSEIF (MOLECULESWITCH == 1 .AND. R_index .EQ. 5) THEN

    ! !  T = 500

       T = 4000. - 12*(time/(24.*3600)-192.)

       IF (T < 500.) T = 500.

       IF (T > 4000.) T = 4000. ! CARE

     CONVERGED = .TRUE.

   END IF

   IONRATES_USE = IONRATES 

   CALL stateq(R_index, N_atoms, atoms, EXCmax, DIMS, x, dep_tot_pervol, IONRATES_USE, fracmat_exc, fground, Nlambda, lambdavec, alfa_c, T, TOTINMAT, Tuse, EXC_MATRIX, convprobs) ! Solve statistical eq. for all atoms for this temperature...solution comes in EXC_MATRIX

   CALL netheating(R_index, heatrate, N_atoms, atoms, EXCmax, EXC_MATRIX, ne, DIMS, x, T, IONRATES_COLLION, nh1, cool_co, cool_sio, fv1_co, fv1_sio, .TRUE.)  ! net heating = heating - cooling at temperature T1

   IF (CONVERGED) EXIT

   T2 = T*t2factor  ! A slightly larger temperature now to compute derivative..

   IONRATES_USE = IONRATES 

   CALL stateq(R_index, N_atoms, atoms, EXCmax, DIMS, x, dep_tot_pervol, IONRATES_USE, fracmat_exc, fground, Nlambda, lambdavec, alfa_c, T2, TOTINMAT, Tuse, EXC_MATRIX2) ! Solve stat eq for temperature T2

   CALL netheating(R_index, heatrate, N_atoms, atoms, EXCmax, EXC_MATRIX2, ne, DIMS, x, T2, IONRATES_COLLION2, nh2, trash1, trash2, trash3, trash4, .FALSE.) ! heating - cooling at temperature T2

   IF (TIMEDEPENDENT_T .EQV. .FALSE.) THEN

        g1 = nh1
        g2 = nh2

   ELSE

        g1 = nh1 - (T-T_prev)/(time-time_prev) - 2*T/time - (T/(1.0+xe))*(xe-xe_prev)/(time-time_prev)
        g2 = nh2 - (T2-T_prev)/(time-time_prev) - 2*T2/time - (T2/(1.0+xe))*(xe-xe_prev)/(time-time_prev)

        !Record above values in a file
        !open(11, file = 'out/plts/timedep_values.dat', status = 'OLD', position='APPEND')
        !write(11,*) nh1, nh2, T, T2, T_prev, xe, xe_prev, time, time_prev
        !close(11)


   END IF

   gprim = (g2-g1)/(T2-T)    ! The derivative of the net heating

   !IF (nh1 > 0 .AND. gprim > 0) THEN  ! If the cooling is insufficient (nh > 0) but still net heating increases with T (hprim > 0), something is wrong..try switching signs to get to a better regime

   !   gprim = -gprim ! CARE

   !END IF

   IF (abs(gprim) > 0) THEN

      !dT = -nh1/hprim             ! The Newton-Raphson step (dx = -f/f')
      dT = -g1/gprim

      IF (dT > 300.) THEN ! Cap T steps to 300 K

         dT = 300.

      ELSEIF (dT < -300.) THEN

         dT = -300.

      END IF

   ELSE

      dT = 0

      CONVERGED = .TRUE.

   END IF

   write(*,'(A15,I8,2F12.3)') 'globid dT T', globid, dT, T

   niterT = niterT + 1   ! Increase no of iternations done by 1 DONT MOVE DOWN, used below

   IF (((TIMEDEPENDENT_T .EQV. .TRUE.) .AND. (ABS(dT/T) < convlimT) .AND. (ABS(g1) < 0.01*heatrate/(3./2.*k*x(DIMS-3)))) .OR. ((TIMEDEPENDENT_T .EQV. .FALSE.) .AND. (ABS(dT/T) < convlimT) .AND. (ABS(nh1) < 0.05*heatrate/(3./2.*k*x(DIMS-3))) .AND. (niterT >=2))) THEN   ! Change in T is small, difference between heating and cooling is small, and at least one iteratin is run --> converge. CARE value Require small change in solution plus a net heating much smaller than the heating

      CONVERGED = .TRUE.

      !T = T + dT  ! Dont do this? Will leave mismatch between T and emissivities

      PRINT '(A20, F10.1, A13, F10.1, A2)', 'T convergence at T= ', T, " K. Last dT: ", dT, " K"

      open(10,file='out/statisticalandthermalequilibrium/td1'//R_index_str(R_index),STATUS='REPLACE')

      !write(10,*) R_index, "T_prev=", T_prev, "time_prev=",time_prev, "T=", T, "time=",time, "nh1=", nh1, "term1a", heatrate/(3./2.*k*x(DIMS-3)), "term1b", -(heatrate/(3./2.*k*x(DIMS-3))-nh1), "term 2", -(T-T_prev)/(time-time_prev), "term 3", -2*T/time

      !write(10,*) heatrate/(3./2.*k*x(DIMS-3)),-(heatrate/(3./2.*k*x(DIMS-3))-nh1), -(T-T_prev)/(time-time_prev), -2*T/time

      apa = heatrate/(3./2.*k*x(DIMS-3))

      !write(10,*) 0, -(heatrate/(3./2.*k*x(DIMS-3))-nh1)/apa,-(T-T_prev)/(time-time_prev)/apa, -2*T/time/apa

      close(10)

      !  print *, 'nh1 heatrate', nh1, heatrate

   ELSE IF (T + dT < 0) THEN   ! T becomes negative --> shrink step

      IF (globid == IDP) PRINT *, 'Shrinking T step..'

      DO WHILE (T + dT < 0)

         dT = 0.1*dT

      END DO

      T = T + dT

   ELSE IF (CONVERGED .EQV. .FALSE.) THEN  ! Ok step, still not converged.

      T = T + dT*damp  ! Take a (damped) NR step

      IONRATES_USE = IONRATES 

      CALL stateq(R_index, N_atoms, atoms, EXCmax, DIMS, x, dep_tot_pervol, IONRATES_USE, fracmat_exc, fground, Nlambda, lambdavec, alfa_c, T, TOTINMAT, Tuse, EXC_MATRIX, convprobs) ! Solve statistical eq. for all atoms for this temperature...solution comes in EXC_MATRIX


   END IF

   IF (T > T_old*1.2) T = T_old *1.2 !CARE not applicable for KNa initial guesses

   IF (T < T_old*0.8) T = T_old*0.8 !CARE not applicable for KNa initial guesses

   !IF (T > 30000) T = 30000 !CARE not applicable for KNa

   IF (T < 10) T = 10 !CARE

   IF (niterT > nitermax) THEN  ! CARE If convergence fails, write warning and use last solution

      PRINT *, 'WARNING, T not converging in statisticalandthermalequilibrium. Proceeding with current value. ', globid

      CONVERGED = .TRUE.  ! CARE

      open(10,file=file_convproblems_suse,STATUS='OLD', POSITION='APPEND')  ! write problem to log

        write(10,*) 'Thermal equilibrium, shell', R_index, T

      close(10)

   END IF

   !SHOULD STRICTLY SPEAKING RECOMPUTE EMISSIVITIES HERE WITH FINAL T..

END DO

open(10,file=TRIM(file_molcool2)//R_index_str(R_index), STATUS='REPLACE')

  write(10,*) R_index, (cool_sio+cool_co)*x(DIMS-8)

close(10)

!open(10,file=TRIM(file_convproblems_suse),STATUS='OLD',POSITION='APPEND')

!DO i = 1, N_atoms

!   IF (convprobs(i)) write(10,*) 'stateq, shell', R_index_str(R_index),'element ', atoms(i)%NAME

!END DO

!close(10)

!open(10,file=TRIM(file_taucont)//R_index_str(R_index), STATUS='REPLACE')  ! Not used for anything

!DO i = 1, Nlambda

!   write(10,'(1000ES11.2)') lambdavec(i), alfa_c(i)*x(DIMS-9)*x(DIMS-12),  alfa_c_ind(i,:)*x(DIMS-9)*x(DIMS-12)  ! DIMS-9 = thickness (cm), DIMS-12 = ff

!END DO

!close(10)

! Temperature and level populations have now converged

adcool = 3*x(DIMS-3)*(1+x(DIMS-2))*k*T/time   ! Adiabatic cooling. 3*ntot*(1+xe)*kT/t  (erg s^-1 cm^-3)........DIMS-3 = ntot, DIMS-2 = xe

open(10,file=TRIM(file_adcool)//R_index_str(R_index), STATUS='REPLACE')

  write(10,*) adcool, adcool/heatrate

close(10)

DO i = 1, N_atoms    ! Compute N_LEVELS_ABSORB (highest level with line tau larger than the treshold)

   atom = atoms(i)

   atoms(i)%N_LEVELS_ABSORB = 0

   N_levels = atom%N_LEVELS

   ALLOCATE(levels_pop(N_levels)); levels_pop = 0

   ALLOCATE(TAU_S(atom%N,N_levels)); TAU_S = 0

   levels_pop = EXC_MATRIX(i,1:N_levels)

   CALL createtau_s(atom, atom%N, N_LEVELS, levels_pop, TAU_S)

   lmax = 0  ! previous max value

   DO l = 1, N_levels-1

      DO u = l+1,atom%N

         IF (ABS(TAU_S(u,l)) > tau_tresh) THEN  ! Allow also negative optical depths to be included

            IF (l > lmax) THEN  ! If this upper level is the highest so far..

               lmax = l  ! replace the record..

               atoms(i)%N_LEVELS_ABSORB = lmax  ! and replace here

            END IF

         END IF

      END DO

   END DO

   DEALLOCATE(levels_pop)

   DEALLOCATE(TAU_S)

END DO

T_eq = T

WRITE (*,'(A10, I5, I10, A20, I10, A10,2F20.4)'), 'Shell T:', R_index, NINT(T_eq), 'last iteration:', NINT(T_old)

END SUBROUTINE statisticalandthermalequilibrium
