SUBROUTINE twophoton(atomname, lambdastart_A, lambdaend_A, stepatstart_A, N, lambdavec, emissvec_integrated)
use PARAMETERS, ONLY : pi, h, c
IMPLICIT NONE
CHARACTER(len=3) :: atomname
DOUBLE PRECISION, INTENT(IN) :: lambdastart_A, lambdaend_A, stepatstart_A
integer, INTENT(OUT) :: N
double precision, DIMENSION(:), ALLOCATABLE,  INTENT(OUT) :: lambdavec, emissvec_integrated
! ==============================================
! TASK :  Computes the two-photon spectrum for HI or HeI [erg s^-1 per atom in the meta-stable state]
! INPUT : atomname      To specify if the HI or HeI twophoton emission is desired. Should be ' HI  ' or 'HeI  '
!         lambastart_A  Lower desired wavelength [A]
!         lambdaend_A   Upper desired wavelength [A]
!         stepatstart_A Step size at lower end [A]
! OUTPUT :N            Number of wavelength bins created (depends on the desired wl range and resolution)
!         lambdavec    WL vector [A]
!         emissvec_integrated   Emission [erg /s / exc state] per bin 
!                       NOTE that the size of the output vectors are only determined after input has been sent.
! GLOBALS :   -
! ==============================================
INTEGER :: i
DOUBLE PRECISION :: dL, y, A, Y1MY, counts
DOUBLE PRECISION, DIMENSION(:), ALLOCATABLE :: emissvec_perA
! -------------------------------------------------------------

y = 0.
Y1MY = 0.

! -------------------------------------------------------------

N = INT(LOG(lambdaend_A/lambdastart_A)/LOG(1.+stepatstart_A/lambdastart_A))  ! Number of bins required to make the logaritmically spaced array

ALLOCATE(lambdavec(N)); lambdavec = 0
ALLOCATE(emissvec_integrated(N)); emissvec_integrated = 0
ALLOCATE(emissvec_perA(N)); emissvec_perA = 0

lambdavec(1) = lambdastart_A

DO i = 2, N   ! Create lambda array, increase relative step size by constant factor
   
   lambdavec(i) = lambdavec(i-1)*(1. + stepatstart_A/lambdastart_A)

END DO

counts = 0

IF (atomname .EQ. 'HI ') THEN  ! HI

   DO i = 1, N-1

      y = 1215.7/lambdavec(i)

      IF (y <= 1) THEN  ! lambda cannot be smaller than 1215.7 A
     
         ! FIT FROM NUSSBAUMER AND SCHMUTZ AA 138, 495 (1984)

500      Y1MY = y*(1.-y)

         A = 202.*(Y1MY * (1. - (4.*Y1MY)**0.8) + 0.88 * Y1MY**1.53 * (4.*Y1MY)**0.8)

         emissvec_perA(i) = 1./(4.*pi)*h*c*1E8/(lambdavec(i))**3.*1215.7*A               ! All lengths in Å..erg /s / A. Note should not be A*y its A(y) in the paper.

         emissvec_integrated(i) = emissvec_perA(i)*(lambdavec(i+1) - lambdavec(i))         ! Total emission in bin...erg s^-1  


         counts = counts + 4*pi*emissvec_integrated(i)/(h*c/(lambdavec(i)*1e-8))

      ELSE

         emissvec_integrated(i) = 0

      END IF

     
   END DO

   emissvec_integrated(N) = 0

ELSE IF (atomname .EQ. 'HeI') THEN  ! HeI

   DO i = 1, N-1

      y = 601.40/lambdavec(i)

      IF (y <= 1) THEN  ! lambda cannot be smaller than 601.4 A

         A = 145*(1-(ABS(1-2*y))**2.5)             ! Li, McCray 1995, ApJ_441_829

         emissvec_perA(i) = 1./(4.*pi)*h*c*1E8/(lambdavec(i))**3.*601.40*A                   ! erg /s/A....All lengths in Å

         emissvec_integrated(i) = emissvec_perA(i)*(lambdavec(i+1) - lambdavec(i))          ! erg/s


      ELSE

         emissvec_integrated(i) = 0

      END IF

   END DO

   emissvec_integrated(N) = 0.
        
ELSE

   WRITE(*,*) 'ERROR in twophoton: Unknown atom.'; STOP 1

END IF
       
!print *, 'counts', counts

END SUBROUTINE twophoton
